# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import os
import stat
import logging
import json
import argparse

from pathlib import Path
from distutils.dir_util import copy_tree

from kqcircuits.simulations.export.util import export_layers
from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.defaults import ELMER_SCRIPT_PATHS
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder


def copy_elmer_scripts_to_directory(path: Path):
    """
    Copies Elmer scripts into directory path.

    Args:
        path: Location where to copy scripts folder.
    """
    if path.exists() and path.is_dir():
        for script_path in ELMER_SCRIPT_PATHS:
            copy_tree(str(script_path), str(path), update=1)


def export_elmer_json(simulation,
                      path: Path,
                      tool='capacitance',
                      linear_system_method='bicgstab',
                      p_element_order=1,
                      frequency=5,
                      mesh_size=None,
                      workflow=None,
                      dielectric_surfaces=None,
                      is_axisymmetric=False):
    """
    Export Elmer simulation into json and gds files.

    Args:
        simulation: The simulation to be exported.
        path: Location where to write json.
        tool(str): Available: "capacitance", "wave_equation" and "cross-section" (Default: capacitance)
        linear_system_method(str): Available: 'bicgstab', 'mg' (Default: bicgstab)
        p_element_order(int): polynomial order of p-elements (Default: 1)
        frequency: Units are in GHz. To set up multifrequency analysis, use list of numbers.
        mesh_size(dict): Parameters to determine mesh element sizes
        workflow(dict): Parameters for simulation workflow
        dielectric_surfaces: Loss tangents for dielectric interfaces, thickness and permittivity should be specified in
            the simulation. The loss tangent is post-processed to the participation to get the quality factor.
            Default is None. Input is of the form::

                'substrate': {
                    'tan_delta_surf': 5e-7
                },
                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                }
        is_axisymmetric(bool): Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)

    Returns:
         Path to exported json file.
    """
    is_cross_section = isinstance(simulation, CrossSectionSimulation)

    if simulation is None or not isinstance(simulation, (Simulation, CrossSectionSimulation)):
        raise ValueError("Cannot export without simulation")

    # collect data for .json file
    if is_cross_section:
        layers = simulation.layer_dict

    json_data = {
        'tool': tool,
        **simulation.get_simulation_data(),
        **({'layers': {k: (v.layer, v.datatype) for k, v in layers.items()}} if is_cross_section else {}),
        'mesh_size': {} if mesh_size is None else mesh_size,
        'workflow': {} if workflow is None else workflow,
        'frequency': frequency,
        **({} if dielectric_surfaces is None else {'dielectric_surfaces': dielectric_surfaces}),
        'linear_system_method': linear_system_method,
        'p_element_order': p_element_order,
        'is_axisymmetric': is_axisymmetric,
    }

    # write .json file
    json_filename = str(path.joinpath(simulation.name + '.json'))
    with open(json_filename, 'w') as fp:
        json.dump(json_data, fp, cls=GeometryJsonEncoder, indent=4)

    # write .gds file
    gds_filename = str(path.joinpath(simulation.name + '.gds'))
    export_layers(gds_filename, simulation.layout, [simulation.cell], output_format='GDS2',
                  layers=layers.values() if is_cross_section else simulation.get_layers())

    return json_filename


def export_elmer_script(json_filenames, path: Path, workflow=None, file_prefix='simulation',
        script_file='scripts/run.py'):
    """
    Create script files for running one or more simulations.
    Create also a main script to launch all the simulations at once.

    Args:
        json_filenames: List of paths to json files to be included into the script.
        path: Location where to write the script file.
        workflow(dict): Parameters for simulation workflow
        file_prefix: Name of the script file to be created.
        script_file: Name of the script file to run.

    Returns:

        Path of exported main script file
    """
    if workflow is None:
        workflow = dict()
    sbatch = 'sbatch_parameters' in workflow
    if sbatch:
        sbatch_parameters = workflow['sbatch_parameters']
        elmer_n_processes = sbatch_parameters['--ntasks']
    python_executable = workflow.get('python_executable', 'python')
    parallelize_workload = 'n_workers' in workflow


    main_script_filename = str(path.joinpath(file_prefix + '.sh'))
    with open(main_script_filename, 'w') as main_file:

        if parallelize_workload and not sbatch:
            main_file.write('{} scripts/simple_workload_manager.py {}'.format(python_executable, workflow['n_workers']))

        script_filenames = []
        for i, json_filename in enumerate(json_filenames):
            with open(json_filename) as f:
                json_data = json.load(f)
                simulation_name = json_data['parameters']['name']
            script_filename = str(path.joinpath(simulation_name + '.sh'))
            script_filenames.append(script_filename)
            with open(script_filename, 'w') as file:
                if sbatch:
                    file.write('#!/bin/bash\n')
                    file.write('#SBATCH --job-name={}_{}\n'.format(sbatch_parameters['--job-name'], str(i)))
                    file.write('#SBATCH --account={}\n'.format(sbatch_parameters['--account']))
                    file.write('#SBATCH --partition={}\n'.format(sbatch_parameters['--partition']))
                    file.write('#SBATCH --time={}\n'.format(sbatch_parameters['--time']))
                    file.write('#SBATCH --ntasks={}\n'.format(elmer_n_processes))
                    file.write('#SBATCH --cpus-per-task={}\n'.format(sbatch_parameters['--cpus-per-task']))
                    file.write('#SBATCH --mem-per-cpu={}\n'.format(sbatch_parameters['--mem-per-cpu']))
                    file.write('\n')
                    file.write('# set the number of threads based on --cpus-per-task\n')
                    file.write('export OMP_NUM_THREADS=$SLURM_CPUS_PER_TASK\n')
                    file.write('\n')
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -n 1 {2} "{0}" "{1}" --only-gmsh -q 2>&1 >> "{1}_Gmsh.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                        )
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write('ElmerGrid 14 2 "{0}" 2>&1 >> "{1}_ElmerGrid.log"\n'.format(
                        simulation_name + ".msh",
                        Path(json_filename).relative_to(path))
                    )
                    file.write('ElmerGrid 2 2 "{0}" 2>&1 -metis {1} 4 --partdual --removeunused >> '\
                            '"{2}_ElmerGrid_part.log"\n'.format(
                        simulation_name,
                        str(elmer_n_processes),
                        Path(json_filename).relative_to(path))
                    )

                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun --cpu-bind none -n {} ElmerSolver_mpi "sif/{}.sif" 2>&1 >> '\
                            '"{}_Elmer.log"\n'.format(
                                elmer_n_processes,
                                simulation_name,
                                Path(json_filename).relative_to(path)
                                ))
                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -n 1 {2} "{0}" "{1}" --write-project-results 2>&1 >> '\
                            '{1}_write_project_results.log\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

                    file.write('echo "Simulation {}/{} write versions json"\n'.format(i + 1, len(json_filenames)))
                    file.write('srun -n 1 {2} "{0}" "{1}" --write-versions-file 2>&1 >> '\
                            '{1}_write_versions_file.log\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                else:
                    file.write('echo "Simulation {}/{} Gmsh"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-gmsh 2>&1 >> "{1}_Gmsh.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} ElmerGrid"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-elmergrid 2>&1 >> "{1}_ElmerGrid.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} Elmer"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-elmer 2>&1 >> "{1}_Elmer.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )
                    file.write('echo "Simulation {}/{} Paraview"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --only-paraview\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

                    file.write('echo "Simulation {}/{} write results json"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --write-project-results 2>&1 >> '\
                            '"{1}_write_project_results.log"\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

                    file.write('echo "Simulation {}/{} write versions file"\n'.format(i + 1, len(json_filenames)))
                    file.write('{2} "{0}" "{1}" --write-versions-file\n'.format(
                        script_file,
                        Path(json_filename).relative_to(path),
                        python_executable)
                    )

            # change permission
            os.chmod(script_filename, os.stat(script_filename).st_mode | stat.S_IEXEC)

            if sbatch:
                main_file.write('echo "Submitting the main script of simulation {}/{}"\n'
                        .format(i + 1, len(json_filenames)))
                main_file.write('echo "--------------------------------------------"\n')
                main_file.write('sbatch "{}"\n'.format(Path(script_filename).relative_to(path)))
            elif parallelize_workload:
                main_file.write(' "./{}"'.format(
                    Path(script_filename).relative_to(path))
                )
            else:
                main_file.write('echo "Submitting the main script of simulation {}/{}"\n'
                        .format(i + 1, len(json_filenames)))
                main_file.write('echo "--------------------------------------------"\n')
                main_file.write('"./{}"\n'.format(Path(script_filename).relative_to(path)))


    # change permission
    os.chmod(main_script_filename, os.stat(main_script_filename).st_mode | stat.S_IEXEC)

    return main_script_filename


def export_elmer(simulations: [],
                 path: Path,
                 tool='capacitance',
                 linear_system_method='bicgstab',
                 p_element_order=3,
                 frequency=5,
                 file_prefix='simulation',
                 script_file='scripts/run.py',
                 mesh_size=None,
                 workflow=None,
                 dielectric_surfaces=None,
                 is_axisymmetric=False,
                 skip_errors=False):
    """
    Exports an elmer simulation model to the simulation path.

    Args:

        simulations(list(Simulation)): list of all the simulations
        path(Path): Location where to output the simulation model
        tool(str): Available: "capacitance", "wave_equation" and "cross-section" (Default: capacitance)
        linear_system_method(str): Available: 'bicgstab', 'mg' (Default: bicgstab)
        p_element_order(int): polynomial order of p-elements (Default: 1)
        frequency: Units are in GHz. To set up multifrequency analysis, use list of numbers.
        file_prefix: File prefix of the script file to be created.
        script_file: Name of the script file to run.
        mesh_size(dict): Parameters to determine mesh element sizes
        workflow(dict): Parameters for simulation workflow
        dielectric_surfaces: Loss tangents for dielectric interfaces, thickness and permittivity should be specified in
            the simulation. The loss tangent is post-processed to the participation to get the quality factor.
            Default is None. Input is of the form::

                'substrate': {
                    'tan_delta_surf': 5e-7
                },
                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                }
        is_axisymmetric(bool): Simulate with Axi Symmetric coordinates along :math:`y\\Big|_{x=0}` (Default: False)
        skip_errors(bool): Skip simulations that cause errors. (Default: False)

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.

    Returns:

        Path to exported script file.
    """
    parser = argparse.ArgumentParser()

    parser.add_argument('-q', "--quiet", action='store_true')
    args, _ = parser.parse_known_args()

    if args.quiet:
        workflow.update(
            {
                'run_gmsh_gui': False,  # For GMSH: if true, the mesh is shown after it is done
                                       # (for large meshes this can take a long time)
                'run_paraview': False,  # this is visual view of the results
            })

    write_commit_reference_file(path)
    copy_elmer_scripts_to_directory(path)
    json_filenames = []
    for simulation in simulations:
        try:
            json_filenames.append(export_elmer_json(simulation=simulation,
                                                    path=path,
                                                    tool=tool,
                                                    linear_system_method=linear_system_method,
                                                    p_element_order=p_element_order,
                                                    frequency=frequency,
                                                    mesh_size=mesh_size,
                                                    workflow=workflow,
                                                    dielectric_surfaces=dielectric_surfaces,
                                                    is_axisymmetric=is_axisymmetric))
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f'Simulation {simulation.name} skipped due to {e.args}. '\
                    'Some of your other simulations might not make sense geometrically. '\
                    'Disable `skip_errors` to see the full traceback.'
                )
            else:
                raise UserWarning(
                    'Generating simulation failed. You can discard the errors using `skip_errors` in `export_elmer`. '\
                    'Moreover, `skip_errors` enables visual inspection of failed and successful simulation '\
                    'geometry files.'
                ) from e

    return export_elmer_script(json_filenames, path, workflow, file_prefix=file_prefix, script_file=script_file)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import point_shift_along_vector
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.airbridges.airbridge import Airbridge


class AirbridgesSim(Simulation):

    n_bridges = Param(pdt.TypeInt, "Number of bridges in series", 5)

    def build(self):
        ab_cell = self.add_element(Airbridge, bridge_length=2*self.b + self.a + 24)
        line_length = 800
        bridge_spacing = (line_length) / (self.n_bridges+1)
        for n in range(self.n_bridges):
            self.insert_cell(ab_cell, pya.DTrans(2, False,
                                                 point_shift_along_vector(pya.DPoint(100, 250), pya.DPoint(150, 250),
                                                                          (n+1) * bridge_spacing)))
        self.produce_waveguide_to_port(pya.DPoint(100, 250), pya.DPoint(150, 250), 1, use_internal_ports=True,
                                       waveguide_length=line_length, term1=6)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.elements.element import Element
from kqcircuits.elements.fluxlines.fluxline import Fluxline
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.junctions.junction import Junction
from kqcircuits.junctions.squid import Squid
from kqcircuits.junctions.sim import Sim


@add_parameters_from(Fluxline, "fluxline_gap_width", "fluxline_type", "fluxline_parameters", "_fluxline_parameters")
@add_parameters_from(Squid, "junction_width", "loop_area", "junction_type",
                     "junction_parameters", "_junction_parameters")
@add_parameters_from(Sim, "junction_total_length")
class Qubit(Element):
    """Base class for qubit objects without actual produce function.

    Collection of shared sub routines for shared parameters and producing shared aspects of qubit geometry including

    * possible fluxlines
    * e-beam layers for SQUIDs
    * SQUID name parameter
    """

    LIBRARY_NAME = "Qubit Library"
    LIBRARY_DESCRIPTION = "Library for qubits."
    LIBRARY_PATH = "qubits"

    mirror_squid =  Param(pdt.TypeBoolean, "Mirror SQUID by its Y axis", False)

    def coerce_parameters_impl(self):
        self.sync_parameters(Fluxline)
        self.sync_parameters(Junction)

    def produce_squid(self, transf, only_arms=False, **parameters):
        """Produces the squid.

        Creates the squid cell and inserts it with the given transformation as a subcell. Also inserts the squid parts
        in "base_metal_gap_wo_grid"-layer to "base_metal_gap_for_EBL"-layer. It also returns a ``right_side`` refpoint,
        calculated from base_metal_gap_wo_grid layer's bounding box to help with arm_length calculation in
        JunctionTestPads.

        Args:
            transf (DCplxTrans): squid transformation
            parameters: other parameters for the squid
            only_arms: Boolean argument that allows to choose whether to create the arms and the squid device or only
                       the arms

        Returns:
            (dict): Relative refpoints for the squid

        """
        cell = self.add_element(Squid, junction_type=self.junction_type, **parameters)
        refpoints_rel = self.get_refpoints(cell)
        mwidth = cell.dbbox_per_layer(self.get_layer("base_metal_gap_wo_grid")).width()
        if mwidth > 0.0:
            refpoints_rel['right_side'] = pya.DPoint(mwidth / 2, 0.0)
        squid_transf = transf * pya.DTrans.M90 if self.mirror_squid else transf

        if "squid_index" in parameters:
            s_index = int(parameters.pop('squid_index'))
            inst, _ = self.insert_cell(cell, squid_transf, inst_name=f"squid_{s_index}")
            inst.set_property("squid_index", s_index)
        else:
            inst, _ = self.insert_cell(cell, squid_transf, inst_name="squid")

        if only_arms:
            # add squid metal etch and unetch shapes to qubit and erase instance of squid
            for layer_name in ["base_metal_gap_wo_grid", "base_metal_addition"]:
                region = pya.Region(cell.shapes(self.get_layer(layer_name)))
                region.transform(inst.cplx_trans)
                self.cell.shapes(self.get_layer(layer_name)).insert(region)
            self.cell.erase(inst)
        else:
            # add parts of qubit to the layer needed for EBL
            region = pya.Region(cell.shapes(self.get_layer("base_metal_gap_wo_grid")))
            region.transform(inst.cplx_trans)
            self.cell.shapes(self.get_layer("base_metal_gap_for_EBL")).insert(region)

        return refpoints_rel

    def produce_fluxline(self, **parameters):
        """Produces the fluxline.

        Creates the fluxline cell and inserts it as a subcell. The "flux" and "flux_corner" ports
        are made available for the qubit.

        Args:
            parameters: parameters for the fluxline to overwrite default and subclass parameters
        """

        if self.fluxline_type == "none":
            return

        cell = self.add_element(Fluxline, **parameters)

        refpoints_so_far = self.get_refpoints(self.cell)
        squid_edge = refpoints_so_far["origin_squid"]
        a = (squid_edge - refpoints_so_far['port_common'])
        rotation = math.atan2(a.y, a.x) / math.pi * 180 + 90
        transf = pya.DCplxTrans(1, rotation, False, squid_edge - self.refpoints["base"])

        cell_inst, _ = self.insert_cell(cell, transf)
        self.copy_port("flux", cell_inst)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.elements.spiral_resonator_polygon import SpiralResonatorPolygon, rectangular_parameters
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.pya_resolver import pya
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.geometry_helper import point_shift_along_vector
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.chips.quality_factor import QualityFactor


@add_parameters_from(QualityFactor, "res_lengths", "n_fingers", "l_fingers", "type_coupler",
                     res_a=[10, 10, 10, 20, 10, 5], res_b=[6, 6, 6, 12, 6, 3])
@add_parameters_from(Chip, frames_enabled=[0, 1])
@add_parameters_from(SpiralResonatorPolygon, "bridge_spacing")
class QualityFactorTwoface(Chip):
    """The PCell declaration for a QualityFactorTwoFace chip.

     Preliminary class for flip-chip resonators.
     """

    resonator_types = Param(pdt.TypeList, "Resonator types (capped, twoface, etched, or solid)", ["capped"] * 6,
                            docstring="Choices: 'capped', 'twoface', 'etched', 'solid'")
    resonator_faces = Param(pdt.TypeList, "Resonator face order list", [0, 1])
    connector_distances = Param(pdt.TypeList, "Resonator input to face to face connector",
                                [500, 1300, 2100, 2900, 3700, 4500], unit="[μm]",
                                docstring="Distances of face to face connectors from resonator inputs")
    spiral_box_height = Param(pdt.TypeDouble, "Spiral resonator box height", 2000)
    spiral_box_width = Param(pdt.TypeDouble, "Spiral resonator box width", 500)
    x_indentation = Param(pdt.TypeDouble, "Resonator/connector indentation from side edges", 800)
    cap_res_distance = Param(pdt.TypeDouble, "Distance between spiral resonator and capacitor", 200)
    waveguide_indentation = Param(pdt.TypeDouble, "Waveguide indentation from top chip edge", 500)
    extra_resonator_avoidance = Param(pdt.TypeList, "Added avoidance", [0, 0, 0, 0, 0, 0], unit="[μm]",
                                      docstring="Added avoidance around resonators [μm]")

    def build(self):
        self._produce_resonators()

    def _produce_resonators(self):
        # Interpretation of parameter lists
        res_lengths = [float(foo) for foo in self.res_lengths]
        res_a = [float(foo) for foo in self.res_a]
        res_b = [float(foo) for foo in self.res_b]
        n_fingers = [float(foo) for foo in self.n_fingers]
        type_coupler = self.type_coupler
        l_fingers = [float(foo) for foo in self.l_fingers]
        connector_distances = [float(foo) for foo in self.connector_distances]
        face1_box = self.get_box(1)
        extra_resonator_avoidance = [float(i) for i in self.extra_resonator_avoidance]

        # Constants
        left_x = face1_box.p1.x + self.waveguide_indentation
        right_x = face1_box.p2.x - self.waveguide_indentation
        left_connector = face1_box.p1.x + self.x_indentation
        right_connector = face1_box.p2.x - self.x_indentation
        mid_y = (face1_box.p1.y + face1_box.p2.y) / 2
        face_config = [self.face_ids[int(i)] for i in self.resonator_faces]

        # Create resonators
        resonators = len(self.res_lengths)
        tl_start = pya.DPoint(left_connector + self.spiral_box_width, mid_y)
        v_res_step = pya.DPoint(right_connector - left_connector - self.spiral_box_width, 0) * \
                     (1. / resonators)
        cell_cross = self.add_element(WaveguideCoplanarSplitter, **t_cross_parameters(
            length_extra_side=5 * self.a_capped, a=self.a_capped, b=self.b_capped, a2=self.a_capped, b2=self.b_capped,
            face_ids=face_config))

        for i in range(resonators):
            # Determine opposite face protection for capacitors and resonators
            res_protect_opposite_face = self.protect_opposite_face
            if self.resonator_types[i] in ["etched", "capped"]:
                res_protect_opposite_face = False
            elif self.resonator_types[i] == "solid":
                res_protect_opposite_face = True

            # Create capacitor
            cplr_params = cap_params(
                n_fingers[i], l_fingers[i], type_coupler[i], protect_opposite_face=res_protect_opposite_face,
                face_ids=face_config, a=res_a[i], b=res_b[i], a2=self.a_capped, b2=self.b_capped)
            cplr = self.add_element(**cplr_params)
            cplr_refpoints_rel = self.get_refpoints(cplr)

            # Every second resonator is on the same side. Define transformations here:
            if i % 2:
                rotation = pya.DTrans.R90 * pya.DTrans.M0
                rot_2 = pya.DTrans.M0
                rot_3 = pya.DTrans.R270
                cplr_pos_post = pya.DTrans.R90 * rot_2 * cplr_refpoints_rel["port_b"]
            else:
                rotation = pya.DTrans.R270
                rot_2 = pya.DTrans.R0
                rot_3 = pya.DTrans.R90
                cplr_pos_post = pya.DTrans.R90 * rot_2 * cplr_refpoints_rel["port_b"] * -1

            # Cross
            cross_trans = pya.DTrans(tl_start + v_res_step * (i + 0.5)) * rot_2
            _, cross_refpoints_abs = self.insert_cell(cell_cross, cross_trans)

            # Coupler
            cplr_pos = cross_refpoints_abs["port_bottom"] + cplr_pos_post
            cplr_trans = pya.DTrans(cplr_pos.x, cplr_pos.y) * rot_3
            inst_cplr, _ = self.insert_cell(cplr, cplr_trans)
            self.refpoints[f'res_{i}_coupler'] = cplr_pos

            # Y-indentation for spiral resonator
            endpoint = point_shift_along_vector(cplr_refpoints_rel["port_b"],
                                                cplr_refpoints_rel["port_b_corner"],
                                                self.cap_res_distance)

            inst_cpw, _ = self.insert_cell(WaveguideCoplanar, **{**self.cell.pcell_parameters_by_name(), **{
                "path": pya.DPath([cplr_refpoints_rel["port_b"], endpoint], 1),
                "term2": 0,
                "a": res_a[i],
                "b": res_b[i],
                "r": self.r,
                "protect_opposite_face": res_protect_opposite_face,
                "face_ids": face_config,
                "margin": self.margin}},
                                           trans=pya.DTrans(cplr_pos) * rotation)

            pos_res_start = cplr_pos - rot_3 * endpoint

            # Spiral resonator
            if self.resonator_types[i] == "twoface":
                res_params = {'connector_dist': connector_distances[i] - self.cap_res_distance, "bridge_spacing": 0}
            elif self.resonator_types[i] == "etched":
                res_params = {
                    'name': 'resonator{}'.format(i),
                    "airbridge_type": "Airbridge Multi Face",
                    "include_bumps": False,
                    "bridge_length": res_a[i] + 2 * (res_b[i] + self.margin),
                    "bridge_width": 2,
                    "pad_length": 2,
                    "bridge_spacing": self.bridge_spacing,
                }
            else:
                res_params = {'name': 'resonator{}'.format(i), "bridge_spacing": 0}
            inst_res, _ = self.insert_cell(SpiralResonatorPolygon,
                                           margin=self.margin + extra_resonator_avoidance[i],
                                           **rectangular_parameters(
                                               right_space=self.spiral_box_height - self.cap_res_distance,
                                               above_space=0,
                                               below_space=self.spiral_box_width,
                                               length=res_lengths[i] - self.cap_res_distance,
                                               a=res_a[i],
                                               b=res_b[i],
                                               face_ids=face_config,
                                               protect_opposite_face=res_protect_opposite_face,
                                               **res_params),
                                           trans=pya.DTrans(pos_res_start) * rotation)

            # Top chip etching and grid avoidance above resonator
            if self.resonator_types[i] == "etched":
                l0 = self.get_layer("ground_grid_avoidance", int(self.resonator_faces[0]))
                region = pya.Region(inst_res.cell.begin_shapes_rec(l0)).transformed(inst_res.trans)
                region += pya.Region(inst_cpw.cell.begin_shapes_rec(l0)).transformed(inst_cpw.trans)
                region += pya.Region(inst_cplr.cell.begin_shapes_rec(l0)).transformed(inst_cplr.trans)
                opposite_face = int(self.resonator_faces[1])
                self.cell.shapes(self.get_layer("ground_grid_avoidance", opposite_face)).insert(region)
                self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", opposite_face)).insert(region)

            # Feedline
            if i == 0:
                left_point = cross_refpoints_abs["port_left"]
            else:
                self.insert_cell(WaveguideCoplanar, **{**self.cell.pcell_parameters_by_name(), **{
                    "path": pya.DPath([right_point] + [cross_refpoints_abs["port_left"]], 1),
                    "term2": 0,
                    "a": self.a_capped,
                    "b": self.b_capped,
                    "r": self.r,
                    "margin": self.margin,
                    "face_ids": face_config
                }})
            right_point = cross_refpoints_abs["port_right"]

        # Launchers
        self.produce_launchers("SMA8", launcher_assignments={8: "PL-IN", 3: "PL-OUT"})

        # Waveguides to the launchers
        if self.waveguide_indentation > 0.0:
            nodes_left = [Node(self.refpoints["PL-IN_port"]),
                          Node((face1_box.p1.x, self.refpoints["PL-IN_port"].y), a=self.a_capped, b=self.b_capped),
                          Node((left_x, self.refpoints["PL-IN_port"].y)),
                          Node((left_x, mid_y)),
                          Node((left_connector, mid_y), face_id=face_config[0]),
                          Node(left_point)]
            nodes_right = [Node(self.refpoints["PL-OUT_port"]),
                           Node((face1_box.p2.x, self.refpoints["PL-OUT_port"].y), a=self.a_capped, b=self.b_capped),
                           Node((right_x, self.refpoints["PL-OUT_port"].y)),
                           Node((right_x, mid_y)),
                           Node((right_connector, mid_y), face_id=face_config[0]),
                           Node(right_point)]
        else:
            nodes_left = [Node(self.refpoints["PL-IN_port"]),
                          Node((left_x, self.refpoints["PL-IN_port"].y)),
                          Node((left_x, mid_y)),
                          Node((face1_box.p1.x, mid_y), a=self.a_capped, b=self.b_capped),
                          Node((left_connector, mid_y), face_id=face_config[0]),
                          Node(left_point)]
            nodes_right = [Node(self.refpoints["PL-OUT_port"]),
                           Node((right_x, self.refpoints["PL-OUT_port"].y)),
                           Node((right_x, mid_y)),
                           Node((face1_box.p2.x, mid_y), a=self.a_capped, b=self.b_capped),
                           Node((right_connector, mid_y), face_id=face_config[0]),
                           Node(right_point)]
        self.insert_cell(WaveguideComposite, nodes=nodes_left)
        self.insert_cell(WaveguideComposite, nodes=nodes_right)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.geometry_helper import circle_polygon, arc_points
from kqcircuits.elements.element import Element
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare


@add_parameters_from(FingerCapacitorSquare, "fixed_length", "a2", "b2")
class CircularCapacitor(Element):
    """The PCell declaration for a circular capacitor.

    An outer semi-circular island with an inside circular island. Fixed-length capacitor is supported and different line
    impedance on each side can be used.
    Two ports with reference points. The feedline has the same length as the width of the ground gap around the coupler.

     .. MARKERS_FOR_PNG 30,0 -56,0 86,0
    """

    r_inner = Param(pdt.TypeDouble, "Internal island radius", 20, unit="μm",
                    docstring="Radius of the outer edge of the center island (μm)")
    r_outer = Param(pdt.TypeDouble, "External island radius, measured at the outer edge", 80, unit="μm",
                    docstring="Radius of the external coupler island (μm)")
    swept_angle = Param(pdt.TypeDouble, "Angle covered by the external island in degrees", 180)
    outer_island_width = Param(pdt.TypeDouble, "External island width", 40, unit="μm",
                               docstring="Width of the external island (μm)")
    ground_gap = Param(pdt.TypeDouble, "Ground plane padding", 20, unit="μm")

    def build(self):
        self.a2 = self.a if self.a2 < 0 else self.a2
        self.b2 = self.b if self.b2 < 0 else self.b2

        y_left = self.a / 2
        y_right = self.a2 / 2
        x_end = self.r_outer + self.ground_gap

        capacitor_region = []
        # generate the inner island
        inner_island = circle_polygon(self.r_inner, self.n)
        capacitor_region.append(inner_island)

        # generate the outer island
        outer_island = self._get_outer_island(self.r_outer, self.outer_island_width, self.swept_angle)
        capacitor_region.append(outer_island)
        capacitor_neg = pya.Region([poly.to_itype(self.layout.dbu) for poly in capacitor_region])

        # add the waveguides inside the ground padding
        capacitor_neg += pya.Region([pya.DPolygon([
            pya.DPoint(-x_end, -y_left),
            pya.DPoint(-0, -y_left),
            pya.DPoint(-0, y_left),
            pya.DPoint(-x_end, y_left),
        ]).to_itype(self.layout.dbu)]) + pya.Region([pya.DPolygon([
            pya.DPoint(x_end, y_right),
            pya.DPoint(self.r_outer - self.outer_island_width / 2, y_right),
            pya.DPoint(self.r_outer - self.outer_island_width / 2, -y_right),
            pya.DPoint(x_end, -y_right),
        ]).to_itype(self.layout.dbu)])
        capacitor_neg.round_corners(5 / self.layout.dbu, 5 / self.layout.dbu, self.n)
        self._add_waveguides(capacitor_neg, x_end, y_left, y_right)

        # define the capacitor in the ground
        ground_region = self._add_ground_region(x_end)
        capacitor_region = ground_region - capacitor_neg
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(capacitor_region)

        # protection region
        region_protection = self._get_protection_region(ground_region)
        self.add_protection(region_protection)

        # ports
        x_port = max(x_end, self.fixed_length / 2)
        self.add_port("a", pya.DPoint(-x_port, 0), pya.DVector(-1, 0))
        self.add_port("b", pya.DPoint(x_port, 0), pya.DVector(1, 0))

        # adds annotation based on refpoints calculated above

    def _get_outer_island(self, r_outer, outer_island_width, swept_angle):
        angle_rad = math.radians(swept_angle)
        points_outside = arc_points(r_outer, -angle_rad / 2, angle_rad / 2, self.n)
        points_inside = arc_points(r_outer - outer_island_width, angle_rad / 2, -angle_rad / 2, self.n)
        points = points_outside + points_inside
        outer_island = pya.DPolygon(points)

        return outer_island

    def _add_ground_region(self, x_end):
        # generate the ground region
        ground_region = []
        island_ground = circle_polygon(self.r_outer + self.ground_gap, self.n)
        ground_region.append(island_ground)
        ground_region = pya.Region([poly.to_itype(self.layout.dbu) for poly in ground_region])
        self._add_waveguides(ground_region, x_end, self.a / 2 + self.b, self.a2 / 2 + self.b2)

        return ground_region

    def _get_protection_region(self, region):
        protection_region = region.sized(self.margin / self.layout.dbu, self.margin / self.layout.dbu, 2)

        return protection_region

    def _add_waveguides(self, region, x_end, y_left, y_right):
        x_guide = self.fixed_length / 2 if (self.fixed_length > 0) else x_end
        if x_guide < x_end:
            raise ValueError(f"Circular capacitor parameters not compatible with fixed_length={self.fixed_length}")
        region += pya.Region([pya.DPolygon([
            pya.DPoint(-x_end + self.margin, -y_left),
            pya.DPoint(-x_guide, -y_left),
            pya.DPoint(-x_guide, y_left),
            pya.DPoint(-x_end + self.margin, y_left),
        ]).to_itype(self.layout.dbu)]) + pya.Region([pya.DPolygon([
            pya.DPoint(x_end - self.margin, y_right),
            pya.DPoint(x_guide, y_right),
            pya.DPoint(x_guide, -y_right),
            pya.DPoint(x_end - self.margin, -y_right),
        ]).to_itype(self.layout.dbu)])



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt


class Launcher(Element):
    """The PCell declaration for a launcher for connecting wirebonds.

    Default wirebond direction to west, waveguide to east. Uses default ratio ``a`` and ``b`` for
    scaling the gap if ``a_launcher`` and ``b_launcher`` are not specified. Taper length is from
    waveguide port to the rectangular part of the launcher pad. Pad width is also used for the
    length of the launcher pad.

     .. MARKERS_FOR_PNG 460,54 410,200 677,46 0,0,300,0
    """

    s = Param(pdt.TypeDouble, "Pad width", 300, unit="μm")
    l = Param(pdt.TypeDouble, "Tapering length", 300, unit="μm")
    a_launcher = Param(pdt.TypeDouble, "Outer trace width", 240, unit="μm")
    b_launcher = Param(pdt.TypeDouble, "Outer gap width", 144, unit="μm")
    launcher_frame_gap = Param(pdt.TypeDouble, "Gap at chip frame", 144, unit="μm")

    def build(self):
        # optical layer

        # shape for the inner conductor
        pts = [
            pya.DPoint(0, self.a / 2 + 0),
            pya.DPoint(self.l, self.a_launcher / 2),
            pya.DPoint(self.l + self.s, self.a_launcher / 2),
            pya.DPoint(self.l + self.s, -self.a_launcher / 2),
            pya.DPoint(self.l, -self.a_launcher / 2),
            pya.DPoint(0, -self.a / 2 + 0)
        ]

        shifts = [
            pya.DVector(0, self.b),
            pya.DVector(0, self.b_launcher),
            pya.DVector(self.launcher_frame_gap, self.b_launcher),
            pya.DVector(self.launcher_frame_gap, -self.b_launcher),
            pya.DVector(0, -self.b_launcher),
            pya.DVector(0, -self.b),
        ]
        pts2 = [p + s for p, s in zip(pts, shifts)]
        pts.reverse()
        shape = pya.DPolygon(pts + pts2)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)

        # protection layer
        shifts = [
            pya.DVector(0, self.margin),
            pya.DVector(0, self.margin),
            pya.DVector(self.margin, self.margin),
            pya.DVector(self.margin, -self.margin),
            pya.DVector(0, -self.margin),
            pya.DVector(0, -self.margin),
        ]
        pts2 = [p + s for p, s in zip(pts2, shifts)]
        shape = pya.DPolygon(pts2)
        self.add_protection(shape)

        # add reference point
        self.add_port("", pya.DPoint(0, 0), pya.DVector(-1, 0))



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


# An example layer config file, compatible with most KQC chips.

from kqcircuits.pya_resolver import pya

default_layers = {}
# face 1 layers
default_layers["1_Nb_gap"] = pya.LayerInfo(0, 0, "1 Nb_gap")
default_layers["1_Nb_gap_avoidance"] = pya.LayerInfo(1, 0, "1 Nb_gap_avoidance")
default_layers["1_Nb_gap_addition"] = pya.LayerInfo(2, 0, "1 Nb_gap_addition")
default_layers["1_Al_1"] = pya.LayerInfo(3, 0, "1 Al_1")
default_layers["1_Al_2"] = pya.LayerInfo(4, 0, "1 Al_2")
default_layers["1_SIS_junction"] = pya.LayerInfo(5, 0, "1 SIS_junction")
default_layers["1_SIS_shadow"] = pya.LayerInfo(6, 0, "1 SIS_shadow")
default_layers["1_underbump"] = pya.LayerInfo(7, 0, "1 underbump")
default_layers["1_new_layer"] = pya.LayerInfo(9, 2, "1 new_layer")  # layer that is not used by KQC elements
default_layers["1_ports"] = pya.LayerInfo(10, 1, "1 ports")
default_layers["1_wg_path"] = pya.LayerInfo(11, 1, "1 wg_path")
# face 2 layers
default_layers["2_Nb_gap"] = pya.LayerInfo(30, 0, "2 Nb_gap")
default_layers["2_Nb_gap_avoidance"] = pya.LayerInfo(31, 0, "2 Nb_gap_avoidance")
default_layers["2_Nb_gap_addition"] = pya.LayerInfo(32, 0, "2 Nb_gap_addition")
default_layers["2_Al_1"] = pya.LayerInfo(33, 0, "2 Al_1")
default_layers["2_Al_2"] = pya.LayerInfo(34, 0, "2 Al_2")
default_layers["2_SIS_junction"] = pya.LayerInfo(35, 0, "2 SIS_junction")
default_layers["2_SIS_shadow"] = pya.LayerInfo(36, 0, "2 SIS_shadow")
default_layers["2_underbump"] = pya.LayerInfo(37, 0, "2 underbump")
default_layers["2_new_layer"] = pya.LayerInfo(39, 2, "2 new_layer")  # layer that is not used by KQC elements
default_layers["2_ports"] = pya.LayerInfo(40, 1, "2 ports")
default_layers["2_wg_path"] = pya.LayerInfo(41, 1, "2 wg_path")
# other layers
default_layers["refpoints"] = pya.LayerInfo(100, 1, "refpoints")
default_layers["waveguide_length"] = pya.LayerInfo(101, 1, "wg_length")
default_layers["annotations"] = pya.LayerInfo(102, 1, "annotations")
default_layers["instance_names"] = pya.LayerInfo(103, 1, "instance_names")


default_faces = {}
default_faces["1"] = {
    "id": "1",
    "base_metal_gap_wo_grid": default_layers["1_Nb_gap"],
    "base_metal_gap_for_EBL": default_layers["1_Nb_gap"],  # same layer can be used for different "face-layers"
    "ground_grid_avoidance": default_layers["1_Nb_gap_avoidance"],
    "base_metal_addition": default_layers["1_Nb_gap_addition"],
    "airbridge_pads": default_layers["1_Al_1"],
    "airbridge_flyover": default_layers["1_Al_2"],
    "SIS_junction": default_layers["1_SIS_junction"],
    "SIS_shadow": default_layers["1_SIS_shadow"],
    "underbump_metallization": default_layers["1_underbump"],
    "ports": default_layers["1_ports"],
    "waveguide_path": default_layers["1_wg_path"],
    "new_layer": default_layers["1_new_layer"],
}
default_faces["2"] = {
    "id": "2",
    "base_metal_gap_wo_grid": default_layers["2_Nb_gap"],
    "base_metal_gap_for_EBL": default_layers["2_Nb_gap"],
    "ground_grid_avoidance": default_layers["2_Nb_gap_avoidance"],
    "base_metal_addition": default_layers["2_Nb_gap_addition"],
    "airbridge_pads": default_layers["2_Al_1"],
    "airbridge_flyover": default_layers["2_Al_2"],
    "SIS_junction": default_layers["2_SIS_junction"],
    "SIS_shadow": default_layers["2_SIS_shadow"],
    "underbump_metallization": default_layers["2_underbump"],
    "ports": default_layers["2_ports"],
    "waveguide_path": default_layers["2_wg_path"],
    "new_layer": default_layers["2_new_layer"],
}

default_face_id = "1"
default_mask_export_layers = None
default_layers_to_mask = None
default_covered_region_excluded_layers = None
mask_bitmap_export_layers = None
all_layers_bitmap_hide_layers = None
chip_export_layer_clusters = None
default_path_length_layers = [
    "1_wg_path",
    "2_wg_path",
    "waveguide_length"
]
default_mask_parameters = {
    "1": {
        "dice_width": 200,
        "text_margin": 100,
    },
    "2": {
        "dice_width": 140,
        "text_margin": 100,
    },
}
default_layer_props = None
default_chip_label_face_prefixes = None
default_parameter_values = {
    "Element": {"face_ids": ["1", "2"]},
}



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.port import InternalPort
from kqcircuits.qubits.double_pads import DoublePads
from kqcircuits.util.parameters import Param, pdt, add_parameters_from

@add_parameters_from(DoublePads)
class DoublePadsSim(Simulation):

    qubit_face = Param(pdt.TypeList, "Bottom or top face qubit position", ['1t1', '2b1'])
    internal_island_ports = Param(pdt.TypeBoolean, "Add InternalPorts on both islands. Use for capacitive simulations.",
                                False)
    junction_inductance = Param(pdt.TypeList, "Qubit junction inductance",
                                11.497e-9, unit="H")
    junction_capacitance = Param(pdt.TypeList, "Qubit junction capacitance",
                                0.1e-15, unit="F")

    def build(self):
        double_pads_cell = self.add_element(DoublePads,
            **{**self.get_parameters(), "junction_type": "Sim", 'face_ids': self.qubit_face})

        qubit_trans = pya.DTrans(0, False, (self.box.left + self.box.right) / 2, (self.box.bottom + self.box.top) / 2)
        _, refp = self.insert_cell(double_pads_cell, qubit_trans, rec_levels=None)

        # translation between elements faces and port faces
        port_face = self.face_ids.index(self.qubit_face[0])

        # Add ports
        port_i = 0
        if self.internal_island_ports:
            self.ports.extend((
                InternalPort((port_i := port_i + 1), refp["port_squid_a"], face=port_face),
                InternalPort((port_i := port_i + 1), refp["port_squid_b"], face=port_face),
            ))
        else:  # Junction between the islands
            self.ports.append(
                InternalPort((port_i := port_i + 1), *self.etched_line(refp["port_squid_a"], refp["port_squid_b"]),
                    face=port_face, inductance=self.junction_inductance, capacitance=self.junction_capacitance,
                    junction=True
                )
            )

        self.produce_waveguide_to_port(refp["port_cplr"], refp["port_cplr_corner"], (port_i := port_i + 1),
            "top", waveguide_length=self.waveguide_length, face=port_face)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node


class AirbridgeCrossings(Chip):
    """The PCell declaration for an AirbridgeCrossings chip.

    On the left side of the chip there is a straight vertical waveguide and a meandering waveguide crossing it multiple
    times. There are airbridges at the crossings. On the right side there is likewise a straight and a meandering
    waveguide, but they do not cross at any point. In the center of the chip there is an array of mechanical tests of
    airbridges with different lengths and widths.
    """

    crossings = Param(pdt.TypeInt, "Number of double crossings", 10,
        docstring="Number of pairs of airbridge crossings")
    b_number = Param(pdt.TypeInt, "Number of bridges", 5,
        docstring="Number of airbridges in one element of the mechanical test array")

    def build(self):

        launchers = self.produce_launchers("SMA8")
        self._produce_transmission_lines(launchers)
        self._produce_mechanical_test_array()

    def _produce_transmission_lines(self, launchers):

        # Left transmission line
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                launchers["NW"][0],
                launchers["SW"][0]
            ], 1)
        )

        # Right transmission line
        self.insert_cell(WaveguideCoplanar,
            path=pya.DPath([
                launchers["NE"][0],
                launchers["SE"][0]
            ], 1)
        )

        # Crossing transmission line
        nodes = [Node(launchers["WN"][0])]
        ref_x = launchers["NW"][0].x
        last_y = launchers["WN"][0].y
        crossings = self.crossings  # must be even
        step = (launchers["WN"][0].y - launchers["WS"][0].y) / (crossings - 0.5) / 2
        wiggle = 250
        for _ in range(crossings):
            nodes.append(Node((ref_x - wiggle, last_y)))
            nodes.append(Node((ref_x, last_y), AirbridgeConnection))
            nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x, last_y), AirbridgeConnection))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
        nodes.append(Node(launchers["WS"][0]))
        waveguide_cell = self.add_element(WaveguideComposite, nodes=nodes)
        self.insert_cell(waveguide_cell)

        # TL without crossings
        nodes = [Node(launchers["EN"][0])]
        ref_x = launchers["NE"][0].x + 2 * wiggle + 50
        last_y = launchers["EN"][0].y
        for _ in range(crossings):
            nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x - wiggle, last_y)))
            nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
        nodes.append(Node(launchers["ES"][0]))
        waveguide_cell = self.add_element(WaveguideComposite, nodes=nodes)
        self.insert_cell(waveguide_cell)

    def _produce_mechanical_test_array(self):

        p_test_origin = pya.DPoint(3600, 9650)
        v_distance_step = pya.DVector(0, -2350)
        v_length_step = pya.DVector(0, -121)
        v_width_step = pya.DVector(400, 0)

        for i, length in enumerate(range(22, 60, 2)):
            for j, width in enumerate(range(5, 20, 2)):
                for k, distance in enumerate(range(2, 22, 5)):
                    loc = p_test_origin + v_length_step * i + v_width_step * j + v_distance_step * k
                    create_airbridges = ((i + k) % 2 == 1)  # airbridges only at every second row
                    self._produce_mechanical_test(loc, distance, self.b_number, length, width, create_airbridges)

    def _produce_mechanical_test(self, loc, distance, number, length, width, create_airbridges):
        # pylint: disable=unused-argument
        wg_len = ((number * (distance + width)) * 2) + 4
        wg_start = loc + pya.DVector(-wg_len / 2, 0)
        wg_end = loc + pya.DVector(+wg_len / 2, 0)
        # v_step = pya.DVector((distance + width) * 2, 0)

        ab = self.add_element(Airbridge,
            # pad_length=1 * width,
            # bridge_length=length,
            # bridge_width=width,
        )
        for i in range(number):
            # ab_trans = pya.DCplxTrans(1, 0, False, wg_start + v_step * (i + 0.5))
            # self.insert_cell(ab, ab_trans)
            if 1000 < loc.y < 9000 and create_airbridges:
                ab_trans = pya.DCplxTrans(1, 0, False, loc + pya.DVector(50*(i - (number-1)/2), 0))
                self.insert_cell(ab, ab_trans)

        # waveguide
        wg = self.add_element(WaveguideCoplanar,
            path=pya.DPath([wg_start, wg_end], 1)
        )
        self.insert_cell(wg)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.elements.flip_chip_connectors import connector_type_choices
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector import FlipChipConnector
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_dc import FlipChipConnectorDc
from kqcircuits.elements.launcher import Launcher
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt


@logged
class FlipChipConnectorRf(FlipChipConnector):
    """PCell declaration for an inter-chip rf connector.

    Flip chip connector with two coplanar waveguide connections and different ground bump configurations.
    The input port is on the first face and on the left. The output port is on the second face and rotated as chosen.

    About connector_type choices:
        * ``Single``: the bump connects the two sides
        * ``GSG``: ground-signal-ground indium bumps
        * ``Coax``: signal transmitting bump is surrounded by four ground bumps

    .. MARKERS_FOR_PNG 0,0 15,0 0,-40 -28,0
    """

    connector_type = Param(pdt.TypeString, "Connector type", "Coax", choices=connector_type_choices)
    inter_bump_distance = Param(pdt.TypeDouble, "Distance between In bumps", 100, unit="μm")
    output_rotation = Param(pdt.TypeDouble, "Rotation of output port w.r.t. input port", 180, unit="degrees")
    connector_a = Param(pdt.TypeDouble, "Conductor width at the connector area", 40, unit="μm")
    connector_b = Param(pdt.TypeDouble, "Gap width at the connector area", 40, unit="μm")
    round_connector = Param(pdt.TypeBoolean, "Use round connector shape", False)
    n_center_bumps = Param(pdt.TypeInt, "Number of center bumps in series", 1)

    def build(self):
        # Flip-chip bump
        bump = self.add_element(FlipChipConnectorDc)
        for i in range(self.n_center_bumps):
            self.insert_cell(bump, pya.DTrans((i - (self.n_center_bumps - 1) / 2) * self.inter_bump_distance, 0))
        bump_ref = self.get_refpoints(bump)
        self.__log.debug("bump_ref: %s", bump_ref)

        tt = pya.DCplxTrans(1, self.output_rotation, False, 0, 0)  # top transformation
        if self.round_connector:
            # Rounded geometry
            def rounded_plate_with_trace(trace_x, plate_x, trace_width, plate_width, plate_length):
                reg = pya.Region(pya.DBox(plate_x - plate_length / 2, -plate_width / 2,
                                          plate_x + plate_length / 2, plate_width / 2).to_itype(self.layout.dbu))
                reg.round_corners(0, min(plate_width, plate_length) / (2 * self.layout.dbu), self.n)
                reg += pya.Region(pya.DBox(trace_x, -trace_width / 2,
                                           plate_x, trace_width / 2).to_itype(self.layout.dbu))
                return reg

            w = self.connector_a + 2 * self.connector_b
            l = w + (self.n_center_bumps - 1) * self.inter_bump_distance
            region = rounded_plate_with_trace(-l/2, 0, self.a + 2 * self.b, w, l)
            avoid_region = region.sized(self.margin / self.layout.dbu, self.margin / self.layout.dbu, 2)
            for i in range(self.n_center_bumps):
                region -= rounded_plate_with_trace(-l/2, (i - (self.n_center_bumps - 1) / 2) * self.inter_bump_distance,
                                                   self.a, self.connector_a, self.connector_a)

            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region)
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", 1)).insert(region.transformed(pya.ICplxTrans(tt)))
            self.add_protection(avoid_region)
            self.add_protection(avoid_region.transformed(pya.ICplxTrans(tt)), 1, 0)

            # add reference point
            self.add_port("{}_port".format(self.face_ids[0]), pya.DPoint(-l/2, 0), pya.DVector(-1, 0), 0)
            self.add_port("{}_port".format(self.face_ids[1]), tt * pya.DPoint(-l/2, 0), tt * pya.DVector(-1, 0), 1)
        else:
            # Taper geometry
            s = self.ubm_diameter + (self.n_center_bumps - 1) * self.inter_bump_distance
            trans = pya.DCplxTrans(1, 0, False, bump_ref["base"] + pya.DPoint(- self.ubm_diameter - s / 2, 0))
            self.insert_cell(Launcher, trans, self.face_ids[0], s=s, l=self.ubm_diameter,
                             a_launcher=self.connector_a, b_launcher=self.connector_b,
                             launcher_frame_gap=self.connector_b)
            self.insert_cell(Launcher, tt * trans, self.face_ids[1], s=s, l=self.ubm_diameter,
                             a_launcher=self.connector_a, b_launcher=self.connector_b,
                             launcher_frame_gap=self.connector_b, face_ids=[self.face_ids[1], self.face_ids[0]])

        # Insert ground bumps
        if self.connector_type == "GSG":
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(0, self.inter_bump_distance)))
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(0, -self.inter_bump_distance)))
        elif self.connector_type == "Coax":
            dist_y = 0.5**0.5 * self.inter_bump_distance
            dist_x = dist_y + self.inter_bump_distance * (self.n_center_bumps - 1) / 2
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(dist_x, dist_y)))
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(-dist_x, dist_y)))
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(dist_x, -dist_y)))
            self.insert_cell(bump, pya.DCplxTrans(1, 0, False, pya.DPoint(-dist_x, -dist_y)))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""Modules for generating and exporting masks.

Masks are represented by mask sets, which can contain one or more mask layouts. Each mask layout consists of a chip
layout and a list of chip variants.

The following things can be exported for the mask and the chips:

    * layout files with specific layers
    * images with specific layers
    * automatic documentation containing:

        - number of each chip in the mask
        - chip parameters
        - images of the mask layout and chips
"""



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.elements.tsvs.tsv import Tsv
from kqcircuits.util.geometry_helper import circle_polygon


class TsvStandard(Tsv):
    """Connector between faces of two sides of a substrate.

    Origin is at the geometric center. Geometry es circular.

    .. MARKERS_FOR_PNG 0,0
    """

    def build(self):
        tsv = circle_polygon(self.tsv_diameter / 2)
        self.cell.shapes(self.get_layer("through_silicon_via")).insert(tsv)
        self.cell.shapes(self.get_layer("through_silicon_via", 1)).insert(tsv)
        margin = circle_polygon(self.tsv_diameter / 2 + self.margin)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(margin)
        self.cell.shapes(self.get_layer("ground_grid_avoidance", 1)).insert(margin)
        del self.refpoints['base']



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.finger_capacitor_taper import FingerCapacitorTaper
from kqcircuits.elements.meander import Meander
from kqcircuits.qubits.swissmon import Swissmon
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar import WaveguideCoplanar
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter, t_cross_parameters
from kqcircuits.test_structures.junction_test_pads.junction_test_pads import JunctionTestPads
from kqcircuits.util.geometry_helper import point_shift_along_vector


@add_parameters_from(Chip, name_chip="Demo")
class Demo(Chip):
    """Demonstration chip with a four qubits, four readout resonators, two probe lines, charge- and fluxlines."""

    readout_res_lengths = Param(pdt.TypeList, "Readout resonator lengths", [5000, 5100, 5200, 5300], unit="[μm]")
    include_couplers = Param(pdt.TypeBoolean, "Include couplers between qubits", True)

    def build(self):

        launcher_assignments = {
            # N
            2: "FL-QB1",
            3: "PL-A-IN",
            4: "PL-A-OUT",
            5: "FL-QB2",
            # E
            7: "DL-QB2",
            12: "DL-QB4",
            # S
            14: "FL-QB4",
            15: "PL-B-IN",
            16: "PL-B-OUT",
            17: "FL-QB3",
            # W
            19: "DL-QB3",
            24: "DL-QB1",
        }
        self.produce_launchers("ARD24", launcher_assignments)

        self.produce_qubits()
        if self.include_couplers:
            self.produce_couplers()
        self.produce_control_lines()
        self.produce_readout_structures()
        self.produce_probelines()
        self.produce_junction_tests()

    def produce_qubits(self):
        dist_x = 3220  # x-distance from chip edge
        dist_y = 3000  # y-distance from chip edge
        self.produce_qubit(pya.DTrans(0, True, dist_x, 1e4 - dist_y), "QB1")
        self.produce_qubit(pya.DTrans(2, False, 1e4 - dist_x, 1e4 - dist_y), "QB2")
        self.produce_qubit(pya.DTrans(0, False, dist_x, dist_y), "QB3")
        self.produce_qubit(pya.DTrans(2, True, 1e4 - dist_x, dist_y), "QB4")

    def produce_qubit(self, trans, inst_name):
        self.insert_cell(Swissmon, trans, inst_name,
                         cpl_length=[120, 120, 120],
                         port_width=[4, 10, 4],
                         )

    def produce_couplers(self):
        self.produce_coupler(1, 2, 2)
        self.produce_coupler(2, 4, 0)
        self.produce_coupler(4, 3, 2)
        self.produce_coupler(3, 1, 0)

    def produce_coupler(self, qubit_a_nr, qubit_b_nr, port_nr):
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["QB{}_port_cplr{}".format(qubit_a_nr, port_nr)]),
            Node(point_shift_along_vector(self.refpoints["QB{}_port_cplr{}".format(qubit_a_nr, port_nr)],
                                          self.refpoints["QB{}_base".format(qubit_a_nr)], -400)),
            Node(point_shift_along_vector(self.refpoints["QB{}_port_cplr{}".format(qubit_b_nr, port_nr)],
                                          self.refpoints["QB{}_base".format(qubit_b_nr)], -400),
                 n_bridges=3
                 ),
            Node(self.refpoints["QB{}_port_cplr{}".format(qubit_b_nr, port_nr)]),
        ], a=4, b=9)

    def produce_control_lines(self):
        for qubit_nr in [1, 2, 3, 4]:
            self.produce_driveline(qubit_nr)
            self.produce_fluxline(qubit_nr)

    def produce_driveline(self, qubit_nr):
        self.insert_cell(
            WaveguideCoplanar,
            path=pya.DPath([
                self.refpoints["DL-QB{}_base".format(qubit_nr)],
                self.refpoints["DL-QB{}_port_corner".format(qubit_nr)],
                self.refpoints["QB{}_port_drive".format(qubit_nr)],
            ], 0),
            term2=self.b
        )

    def produce_fluxline(self, qubit_nr):
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["FL-QB{}_base".format(qubit_nr)]),
            Node(self.refpoints["FL-QB{}_port_corner".format(qubit_nr)]),
            Node(self.refpoints["QB{}_port_flux_corner".format(qubit_nr)], n_bridges=4),
            Node(self.refpoints["QB{}_port_flux".format(qubit_nr)])
        ])

    def produce_readout_structures(self):
        self.produce_readout_structure(1, False, 4)
        self.produce_readout_structure(2, True, 8)
        self.produce_readout_structure(3, False, 5)
        self.produce_readout_structure(4, True, 7)

    def produce_readout_structure(self, qubit_nr, mirrored, cap_finger_nr):

        # non-meandering part of the resonator

        point_1 = self.refpoints["QB{}_port_cplr1".format(qubit_nr)]
        point_2 = point_shift_along_vector(self.refpoints["QB{}_port_cplr1".format(qubit_nr)],
                                           self.refpoints["QB{}_base".format(qubit_nr)], -600)
        point_3 = point_2 + pya.DPoint(-300 if mirrored else 300, 0)

        waveguide_inst, _ = self.insert_cell(WaveguideComposite, nodes=[
            Node(point_1),
            Node(point_2),
            Node(point_3),
        ])
        length_nonmeander = waveguide_inst.cell.length()

        # meandering part of the resonator

        meander_start = point_3
        meander_end = point_3 + pya.DPoint(-1100 if mirrored else 1100, 0)

        self.insert_cell(Meander,
                         start=meander_start,
                         end=meander_end,
                         length=float(self.readout_res_lengths[qubit_nr - 1]) - length_nonmeander,
                         )

        # capacitor and tcross waveguide connecting resonator to probeline

        if mirrored:
            cap_rot = 2
            tcross_rot = 1
        else:
            cap_rot = 0
            tcross_rot = 3

        _, cap_ref_abs = self.insert_cell(FingerCapacitorSquare, pya.DTrans(cap_rot), align_to=meander_end,
                                          align="port_a", finger_number=cap_finger_nr)

        self.insert_cell(WaveguideCoplanarSplitter, pya.DTrans(tcross_rot, False, 0, 0),
                         inst_name="PL{}".format(qubit_nr),
                         label_trans=pya.DCplxTrans(0.2), align_to=cap_ref_abs["port_b"], align="port_bottom",
                         **t_cross_parameters(a=self.a, b=self.b, a2=self.a, b2=self.b, length_extra_side=30))

    def produce_probelines(self):
        self.produce_probeline("PL-A", 1, 2, False, 6)
        self.produce_probeline("PL-B", 4, 3, True, 3)

    def produce_probeline(self, probeline_name, qubit_a_nr, qubit_b_nr, mirrored, cap_finger_nr):

        if mirrored:
            cap_rot = 1
            cap_pos_shift = pya.DPoint(0, -2000)
        else:
            cap_rot = 3
            cap_pos_shift = pya.DPoint(0, 2000)

        cap_trans = pya.DTrans(cap_rot, False, self.refpoints["PL{}_port_left".format(qubit_a_nr)] + cap_pos_shift)
        _, cap_ref_abs = self.insert_cell(FingerCapacitorTaper, cap_trans, finger_number=cap_finger_nr, taper_length=20)

        # segment 1
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["{}-IN_base".format(probeline_name)]),
            Node(self.refpoints["{}-IN_port_corner".format(probeline_name)]),
            Node((self.refpoints["PL{}_port_left".format(qubit_a_nr)].x,
                  self.refpoints["{}-IN_port_corner".format(probeline_name)].y)),
            Node(cap_ref_abs["port_a"]),
        ])

        # segment 2
        self.insert_cell(WaveguideComposite, nodes=[
            Node(cap_ref_abs["port_b"]),
            Node((self.refpoints["PL{}_port_left".format(qubit_a_nr)].x,
                  self.refpoints["QB{}_base".format(qubit_a_nr)].y),
                 AirbridgeConnection if self.include_couplers else None
                 ),
            Node(self.refpoints["PL{}_port_left".format(qubit_a_nr)]),
        ])

        # segment 3
        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["PL{}_port_right".format(qubit_a_nr)]),
            Node(point_shift_along_vector(self.refpoints["PL{}_port_right".format(qubit_a_nr)],
                                          self.refpoints["PL{}_port_right_corner".format(qubit_a_nr)], 600)),
            Node(point_shift_along_vector(self.refpoints["PL{}_port_left".format(qubit_b_nr)],
                                          self.refpoints["PL{}_port_left_corner".format(qubit_b_nr)], 600),
                 n_bridges=1
                 ),
            Node(self.refpoints["PL{}_port_left".format(qubit_b_nr)]),
        ])

        self.insert_cell(WaveguideComposite, nodes=[
            Node(self.refpoints["{}-OUT_base".format(probeline_name)]),
            Node(self.refpoints["{}-OUT_port_corner".format(probeline_name)]),
            Node((self.refpoints["PL{}_port_right".format(qubit_b_nr)].x,
                  self.refpoints["{}-OUT_port_corner".format(probeline_name)].y)),
            Node((self.refpoints["PL{}_port_right".format(qubit_b_nr)].x,
                  self.refpoints["QB{}_base".format(qubit_b_nr)].y),
                 AirbridgeConnection if self.include_couplers else None
                 ),
            Node(self.refpoints["PL{}_port_right".format(qubit_b_nr)]),
        ])

    def produce_junction_tests(self):
        junction_test_cell = self.add_element(JunctionTestPads, margin=50, area_height=2500)
        label_trans = pya.DCplxTrans(0.5)
        self.insert_cell(junction_test_cell, pya.DTrans(0, False, 900, 3750), "testarray_w", label_trans=label_trans)
        self.insert_cell(junction_test_cell, pya.DTrans(0, False, 7800, 3750), "testarray_e", label_trans=label_trans)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from kqcircuits.pya_resolver import pya


class Refpoints:
    """Helper class for extracting reference points from given layer and cell.

    Once Refpoints is initialized, it can be used similar way as dictionary, where reference point text (string) field
    is the key and reference point position (pya.DPoint) is the value.

    Refpoints is implemented such that the dictionary is extracted from given layer and cell only when it's used for the
    first time. Extracting the dictionary can be relatively time-demanding process, so this way we can speed up the
    element creation process in KQC.

    Attributes:
        layer: layer specification for source of reference points
        cell: cell containing the reference points
        trans: transform for converting reference points into target coordinate system
        rec_levels: recursion level when looking for reference points from subcells. Set to 0 to disable recursion.
    """
    def __init__(self, layer, cell, trans, rec_levels):
        self.layer = layer
        self.cell = cell
        self.trans = trans
        self.rec_levels = rec_levels
        self.refpoints = None

    def dict(self):
        """Extracts and returns reference points as dictionary, where text is the key and position is the value."""
        if self.refpoints is None:
            self.refpoints = {}
            shapes_iter = pya.RecursiveShapeIterator(self.cell.layout(), self.cell, self.layer)
            if self.rec_levels is not None:
                shapes_iter.max_depth = self.rec_levels
            while not shapes_iter.at_end():
                shape = shapes_iter.shape()
                if shape.type() in (pya.Shape.TText, pya.Shape.TTextRef):
                    self.refpoints[shape.text_string] = self.trans * (shapes_iter.dtrans()*pya.DPoint(shape.text_dpos))
                shapes_iter.next()
        return self.refpoints

    def __iter__(self):
        """Returns iterator"""
        return iter(self.dict())

    def __getitem__(self, item):
        """The [] operator to return position for given reference point text."""
        return self.dict()[item]

    def __setitem__(self, item, value):
        """The [] operator to set a new reference point."""
        self.dict()[item] = value

    def items(self):
        """Returns a list of text-position pairs."""
        return self.dict().items()

    def keys(self):
        """Returns a list of texts."""
        return self.dict().keys()

    def values(self):
        """Returns a list of positions."""
        return self.dict().values()



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import math
from autologging import logged
from tqdm import tqdm

from kqcircuits.pya_resolver import pya
from kqcircuits.defaults import default_layers, default_brand, default_faces, default_mask_parameters, \
    default_layers_to_mask, default_covered_region_excluded_layers, default_mask_export_layers, default_bar_format
from kqcircuits.elements.markers.marker import Marker
from kqcircuits.elements.mask_marker_fc import MaskMarkerFc
from kqcircuits.util.label import produce_label, LabelOrigin
from kqcircuits.util.merge import merge_layout_layers_on_face, convert_child_instances_to_static


@logged
class MaskLayout:
    """Class representing the mask for a certain face.

    A MaskLayout is used to create the cells for the mask.

    Attributes:
        layout: pya.Layout for this mask
        name: Name of the mask
        version: Mask version
        with_grid: Boolean determining if ground grid is generated
        face_id: face_id of this mask layout, "1t1" | "2b1" | "2t1"
        layers_to_mask: dictionary of layers with mask label postfix for mask label and mask covered region creation
        covered_region_excluded_layers: list of layers in `layers_to_mask` for which mask covered region is not created
        chips_map: List of lists (2D-array) of strings, each string is a chip name (or --- for no chip)
        chips_map_legend: Dictionary where keys are chip names, values are chip cells
        wafer_rad: Wafer radius
        wafer_center: Wafer center as a pya.DVector
        chips_map_offset: Offset to make chips_map centered on wafer
        wafer_top_flat_length: length of flat edge at the top of the wafer
        wafer_bottom_flat_length: length of flat edge at the bottom of the wafer
        dice_width: Dicing width for this mask layout
        text_margin: Text margin for this mask layout
        chip_size: side width of the chips (assuming square chips)
        edge_clearance: minimum clearance of outer chips from the edge of the mask, measured from the chip center
        chip_box_offset: Offset (pya.DVector) from chip origin of the chip frame boxes for this face
        chip_trans: DTrans applied to all chips
        mask_name_offset: mask name label offset from default position (DPoint)
        mask_name_scale: text scaling factor for mask name label (float)
        mask_text_scale: text scaling factor for graphical representation layer (float)
        mask_marker_offset: offset of mask markers from wafer center in horizontal and vertical directions (float)
        mask_export_layers: list of layer names (without face_ids) to be exported as individual mask `.oas` files
        submasks: list of submasks, each element is a tuple (submask mask_layout, submask position)
        extra_id: extra string used to create unique name for mask layouts with the same face_id
        extra_chips: List of tuples (name, position, trans) for chips placed outside chips_map, trans is an optional
            transformation to use in place of self.chip_trans
        top_cell: Top cell of this mask layout
        added_chips: List of (chip name, chip position, chip bounding box, chip dtrans) populated by chips added during
            build()
    """

    def __init__(self, layout, name, version, with_grid, chips_map, face_id, **kwargs):

        self.layout: pya.Layout = layout
        self.name = name
        self.version = version
        self.with_grid = with_grid
        self.face_id = face_id
        self.chips_map = chips_map
        self.chips_map_legend = None

        self.layers_to_mask = kwargs.get("layers_to_mask", default_layers_to_mask)
        self.covered_region_excluded_layers = kwargs.get("covered_region_excluded_layers",
                                                        default_covered_region_excluded_layers) + ["mask_edge"]
        self.wafer_rad = kwargs.get("wafer_rad", default_mask_parameters[self.face_id]["wafer_rad"])
        self.wafer_center = (pya.DVector(0, 0))
        self.chips_map_offset = kwargs.get("chips_map_offset",
                                              default_mask_parameters[self.face_id]["chips_map_offset"])
        self.wafer_top_flat_length = kwargs.get("wafer_top_flat_length", 0)
        self.wafer_bottom_flat_length = kwargs.get("wafer_bottom_flat_length", 0)
        self.dice_width = kwargs.get("dice_width", default_mask_parameters[self.face_id]["dice_width"])
        self.text_margin = kwargs.get("text_margin", default_mask_parameters[self.face_id]["text_margin"])
        self.chip_size = kwargs.get("chip_size", default_mask_parameters[self.face_id]["chip_size"])
        self.edge_clearance = kwargs.get("edge_clearance", self.chip_size)
        self.chip_box_offset = kwargs.get("chip_box_offset", default_mask_parameters[self.face_id]["chip_box_offset"])
        self.chip_trans = kwargs.get("chip_trans", default_mask_parameters[self.face_id]["chip_trans"])
        self.mask_name_offset = kwargs.get("mask_name_offset", default_mask_parameters[self.face_id][
            "mask_name_offset"])
        self.mask_name_scale = kwargs.get("mask_name_scale", 1)
        self.mask_text_scale = kwargs.get("mask_text_scale", default_mask_parameters[self.face_id]["mask_text_scale"])
        self.mask_markers_type = kwargs.get("mask_markers_type", "all")
        self.mask_marker_offset = kwargs.get("mask_marker_offset", default_mask_parameters[self.face_id][
            "mask_marker_offset"])
        self.mask_export_layers = kwargs.get("mask_export_layers", default_mask_export_layers)
        self.submasks = kwargs.get("submasks", [])
        self.extra_id = kwargs.get("extra_id", "")
        self.extra_chips = kwargs.get("extra_chips", [])

        self.top_cell = self.layout.create_cell(f"{self.name} {self.face_id}")
        self.added_chips = []

    def build(self, chips_map_legend):
        """Builds the cell hierarchy for this mask layout.

        Inserts cells copied from chips_map_legend to self.top_cell at positions determined by self.chips_map. The
        copied cells are modified to only have layers corresponding to self.face_id, and they are translated and/or
        mirrored correctly based on self.face_id. Also inserts cells for mask markers, mask name label, and the circular
        area covered by the mask.

        Args:
            chips_map_legend: Dictionary where keys are chip names, values are chip cells

        """
        self.chips_map_legend = {}

        for name, cell in tqdm(chips_map_legend.items(), desc='Building cell hierarchy', bar_format=default_bar_format):

            # pylint: disable=use-a-generator
            if any([name in row for row in self.chips_map] + [chip[0] == name for chip in self.extra_chips]):

                # create copies of the chips, so that modifying these only affects the ones in this MaskLayout
                new_cell = self.layout.create_cell(name)
                new_cell.copy_tree(cell)
                # remove layers belonging to another face
                for face_id, face_dictionary in default_faces.items():
                    if face_id != self.face_id:
                        for layer_info in face_dictionary.values():
                            shapes_iter = new_cell.begin_shapes_rec(self.layout.layer(layer_info))

                            # iterating shapes using shapes_iter.at_end() fails:
                            # https://www.klayout.de/forum/discussion/comment/4275
                            # solution is to use a separate buffer list to iterate
                            shapes = []
                            while not shapes_iter.at_end():
                                try:
                                    shapes.append(shapes_iter.shape())
                                    shapes_iter.next()
                                except ValueError:
                                    print("error occurs at %s at %s" % (name, face_id))

                            for shapes_to_remove in shapes:
                                shapes_to_remove.delete()

                self.chips_map_legend[name] = new_cell

        step_ver = pya.DVector(0, -self.chip_size)
        step_hor = pya.DVector(self.chip_size, 0)

        region_covered = self._mask_create_geometry()
        if len(self.submasks) > 0:
            region_covered = pya.Region()  # don't fill with metal gap layer if using submasks
            for submask_layout, submask_pos in self.submasks:
                self.top_cell.insert(
                    pya.DCellInstArray(submask_layout.top_cell.cell_index(),
                                       pya.DTrans(submask_pos - submask_layout.wafer_center + self.wafer_center))
                )

        # add chips from chips_map
        for (i, row) in enumerate(tqdm(self.chips_map, desc='Adding chips to mask', bar_format=default_bar_format)):
            for (j, chip_name) in enumerate(row):
                position = pya.DPoint(step_ver * (i + 1) + step_hor * j) - self.chips_map_offset \
                           + pya.DVector(-self.wafer_rad, self.wafer_rad)
                if (position - step_ver*0.5 + step_hor*0.5 - self.wafer_center).abs() - self.wafer_rad < \
                        -self.edge_clearance:
                    added_chip, region_chip = self._add_chip(chip_name, position, self.chip_trans)
                else:
                    added_chip, region_chip = False, pya.Region()
                region_covered -= region_chip
                if not added_chip:
                    self.chips_map[i][j] = "---"

        # add chips outside chips_map
        for name, pos, *trans in self.extra_chips:  # trans is optional
            if name in chips_map_legend:
                region_covered -= self._add_chip(name, pos, trans[0] if trans else self.chip_trans)[1]
                self.chips_map.append([name])  # to get correct amount of chips in mask documentation

        maskextra_cell: pya.Cell = self.layout.create_cell("MaskExtra")

        self._insert_mask_name_label(self.top_cell, default_layers["mask_graphical_rep"])
        self._mask_create_covered_region(maskextra_cell, region_covered, self.layers_to_mask)
        convert_child_instances_to_static(self.layout, maskextra_cell, only_elements=True, prune=True)
        merge_layout_layers_on_face(self.layout, maskextra_cell, self.face())

    def insert_chip_copy_labels(self, labels_cell, layers):
        """Inserts chip copy labels to all chips in this mask layout and its submasks

        Args:
            labels_cell: Cell to which the labels are inserted
            layers: list of layer names (without face_ids) where the labels are produced
        """

        # find labels_cell for this mask and each submask
        labels_cells = {self: labels_cell}
        for submask_layout, submask_pos in self.submasks:
            for inst in submask_layout.top_cell.each_inst():
                # workaround for getting the cell due to KLayout bug, see
                # https://www.klayout.de/forum/discussion/1191
                # TODO: replace by `inst_cell = inst.cell` once KLayout bug is fixed
                inst_cell = submask_layout.layout.cell(inst.cell_index)
                if inst_cell.name.startswith("ChipLabels"):
                    labels_cells[submask_layout] = inst_cell
                    break

        # find all unique x and y coords of chips and place them in the corresponding keys in chips_dict
        chips_dict = {}  # {(pos_x, pos_y): chip_name, chip_pos (in submask coordinates), chip_inst, mask_layout}
        xvals = set()
        yvals = set()
        for chip_name, pos, bbox, dtrans in self.added_chips:
            xvals.add(pos.x)
            yvals.add(pos.y)
            chips_dict[(pos.x, pos.y)] = chip_name, pos, bbox, dtrans, self
        for submask_layout, submask_pos in self.submasks:
            for chip_name, pos, bbox, dtrans in submask_layout.added_chips:
                pos2 = pos + submask_pos
                xvals.add(pos2.x)
                yvals.add(pos2.y)
                chips_dict[(pos2.x, pos2.y)] = chip_name, pos, bbox, dtrans, submask_layout

        # produce the labels such that chips with identical x-coordinate (y-coordinate) have identical number (letter)
        for i, y in enumerate(sorted(yvals, reverse=True)):
            for j, x in enumerate(sorted(xvals)):
                if (x, y) in chips_dict:
                    chip_name, _, bbox, dtrans, mask_layout = chips_dict[(x, y)]
                    labels_cell_2 = labels_cells[mask_layout]
                    pos_index_name = chr(ord("A") + i) + ("{:02d}".format(j))
                    bbox_x1 = bbox.left if dtrans.is_mirror() else bbox.right
                    produce_label(labels_cell_2, pos_index_name, dtrans*(pya.DPoint(bbox_x1, bbox.bottom)),
                                  LabelOrigin.BOTTOMRIGHT, mask_layout.dice_width, mask_layout.text_margin,
                                  [mask_layout.face()[layer] for layer in layers],
                                  mask_layout.face()["ground_grid_avoidance"])
                    bbox_x2 = bbox.right if dtrans.is_mirror() else bbox.left
                    mask_layout._add_chip_graphical_representation_layer(chip_name,
                                                                         dtrans*(pya.DPoint(bbox_x2, bbox.bottom)),
                                                                         pos_index_name, bbox.width(), labels_cell_2)

    def face(self):
        """Returns the face dictionary for this mask layout"""
        return default_faces[self.face_id]

    def _mask_create_geometry(self):
        y_clip = -14.5e4

        points = []
        for a in range(0, 256 + 1):
            x = math.cos(a / 128 * math.pi) * self.wafer_rad
            y = max(math.sin(a / 128 * math.pi) * self.wafer_rad, y_clip)
            if (y > 0 and (x > self.wafer_top_flat_length/2 or x < -self.wafer_top_flat_length/2)) or \
               (y < 0 and (x > self.wafer_bottom_flat_length/2 or x < -self.wafer_bottom_flat_length/2)):
                points.append(pya.DPoint(self.wafer_center.x + x, self.wafer_center.y + y))

        region_covered = pya.Region(pya.DPolygon(points).to_itype(self.layout.dbu))
        return region_covered

    def _add_chip(self, name, position, trans):
        """Returns a tuple (Boolean telling if the chip was added, Region which the chip covers)."""
        chip_region = pya.Region()
        if name in self.chips_map_legend.keys():
            chip_cell, bounding_box, bbox_offset = self._get_chip_cell_and_bbox(name)
            trans = pya.DTrans(position + pya.DVector(bbox_offset, 0) - self.chip_box_offset)*trans
            self.top_cell.insert(pya.DCellInstArray(chip_cell.cell_index(), trans))
            chip_region = pya.Region(pya.Box(trans*bounding_box*(1/self.layout.dbu)))
            self.added_chips.append((name, position, bounding_box, trans))
            return True, chip_region
        return False, chip_region

    def _mask_create_covered_region(self, maskextra_cell, region_covered, layers_dict):
        dbu = self.layout.dbu

        leftmost_label_x = 1e10
        labels = []
        for layer, postfix in layers_dict.items():
            label_cell, label_trans = self._create_mask_name_label(self.face()[layer], postfix)
            if label_trans.disp.x < leftmost_label_x:
                leftmost_label_x = label_trans.disp.x
            labels.append((label_cell, label_trans))
        # align left edges of mask name labels in different layers
        for (label_cell, label_trans) in labels:
            label_trans = pya.DTrans(label_trans.rot, label_trans.is_mirror(), leftmost_label_x, label_trans.disp.y)
            inst = maskextra_cell.insert(pya.DCellInstArray(label_cell.cell_index(), label_trans))
            region_covered -= pya.Region(inst.bbox()).extents(1e3 / dbu)

        circle = pya.DTrans(self.wafer_center) * pya.DPath(
            [pya.DPoint(math.cos(a / 32 * math.pi) * self.wafer_rad, math.sin(a / 32 * math.pi) * self.wafer_rad)
             for a in range(0, 64 + 1)], 100)
        maskextra_cell.shapes(self.layout.layer(default_layers["mask_graphical_rep"])).insert(circle)

        offset = self.mask_marker_offset
        # Corner mask markers
        if self.mask_markers_type in ["all", "only_corners"]:
            cell_marker = Marker.create(self.layout, window=True, face_ids=[self.face_id])
            marker_transes = [pya.DTrans(self.wafer_center.x - offset, self.wafer_center.y - offset) * pya.DTrans.R180,
                              pya.DTrans(self.wafer_center.x + offset, self.wafer_center.y - offset) * pya.DTrans.R270,
                              pya.DTrans(self.wafer_center.x - offset, self.wafer_center.y + offset) * pya.DTrans.R90,
                              pya.DTrans(self.wafer_center.x + offset, self.wafer_center.y + offset) * pya.DTrans.R0]
            self._add_markers(maskextra_cell, region_covered, cell_marker, marker_transes)
        # Side mask markers
        if self.mask_markers_type in ["all", "only_sides"]:
            cell_marker = MaskMarkerFc.create(self.layout,  window=True, face_ids=[self.face_id])
            marker_transes = [
                              pya.DTrans(self.wafer_center.x - offset * 1.9**0.5, self.wafer_center.y) * pya.DTrans.M90,
                              pya.DTrans(self.wafer_center.x + offset * 1.9**0.5, self.wafer_center.y) * pya.DTrans.R0]
            self._add_markers(maskextra_cell, region_covered, cell_marker, marker_transes)

        maskextra_cell.shapes(self.layout.layer(default_layers["mask_graphical_rep"])).insert(region_covered)
        # remove unwanted circle boundary and filling from `layers_to_mask` which have been excluded in
        # `covered_region_excluded_layers`
        for layer_name in layers_dict.keys():
            if layer_name not in self.covered_region_excluded_layers:
                maskextra_cell.shapes(self.layout.layer(self.face()[layer_name])).insert(region_covered)

        self.top_cell.insert(pya.DCellInstArray(maskextra_cell.cell_index(), pya.DTrans()))

    def _add_markers(self, maskextra_cell, region_covered, cell_marker, marker_transes):
        for trans in marker_transes:
            inst = maskextra_cell.insert(pya.DCellInstArray(cell_marker.cell_index(), trans))
            region_covered -= pya.Region(inst.bbox()).extents(1e3/self.layout.dbu)

    def _get_chip_name(self, search_cell):
        for chip_name, cell in self.chips_map_legend.items():
            if search_cell == cell:
                return chip_name
        return ""

    def _get_chip_cell_and_bbox(self, chip_name):
        chip_cell = self.chips_map_legend[chip_name]
        bounding_box = chip_cell.dbbox_per_layer(self.layout.layer(self.face()["base_metal_gap_wo_grid"]))
        bbox_offset = self.chip_size - bounding_box.width()  # for chips that are smaller than self.chip_size
        return chip_cell, bounding_box, bbox_offset

    def _add_chip_graphical_representation_layer(self, chip_name, position, pos_index_name, chip_size, cell):
        chip_name_text = self.layout.create_cell("TEXT", "Basic", {
            "layer": default_layers["mask_graphical_rep"],
            "text": chip_name,
            "mag": 15000 * self.mask_text_scale / len(chip_name),
        })
        pos_index_name_text = self.layout.create_cell("TEXT", "Basic", {
            "layer": default_layers["mask_graphical_rep"],
            "text": pos_index_name,
            "mag": 4000 * self.mask_text_scale,
        })
        chip_name_trans = pya.DTrans(position + pya.DVector((chip_size - chip_name_text.dbbox().width()) / 2,
                                                            self.mask_text_scale * 750))
        cell.insert(pya.DCellInstArray(chip_name_text.cell_index(), chip_name_trans))
        pos_index_trans = pya.DTrans(position + pya.DVector((chip_size - pos_index_name_text.dbbox().width()) / 2,
                                                            self.mask_text_scale * 6000))
        cell.insert(pya.DCellInstArray(pos_index_name_text.cell_index(), pos_index_trans))

    def _insert_mask_name_label(self, cell, layer, postfix=""):
        cell_mask_name, trans = self._create_mask_name_label(layer, postfix)
        inst = cell.insert(pya.DCellInstArray(cell_mask_name.cell_index(), trans))
        return inst

    def _create_mask_name_label(self, layer, postfix=""):
        if postfix != "":
            postfix = "-" + postfix
        cell_mask_name = self.layout.create_cell("TEXT", "Basic", {
            "layer": layer,
            "text": default_brand + "-" + self.name + "v" + str(self.version) + postfix,
            "mag": self.mask_name_scale*5000.0,
        })
        cell_mask_name_h = cell_mask_name.dbbox().height()
        cell_mask_name_w = cell_mask_name.dbbox().width()
        trans = pya.DTrans(self.wafer_center.x + self.mask_name_offset.x - cell_mask_name_w / 2,
                           self.wafer_rad + self.mask_name_offset.y - cell_mask_name_h / 2)

        return cell_mask_name, trans



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.chip import Chip
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from

@add_parameters_from(Chip, frames_enabled=[0, 1])
class CrossingTwoface(Chip):
    """The PCell declaration for a CrossingTwoFace chip.

    On the left side of the chip there is a straight vertical waveguide bottom face and a meandering waveguide crossing
    multiple times on the top face. There are transmission lines at different faces at the crossings. On the right side
    there is likewise a straight and a meandering waveguide, but they do not cross at any point.
    """

    crossings = Param(pdt.TypeInt, "Number of double crossings", 3,
        docstring="Number of pairs of flip-chip crossings")
    crossing_length = Param(pdt.TypeDouble, "Crossing waveguide length", 400, unit="μm",
        docstring="Length of the crossing on the top face (μm)")
    cross_talk_distance = Param(pdt.TypeDouble, "Transmission line distance from meander", 300, unit="μm",
        docstring="Distance between the right straight transmission line and meander on the right (μm)")
    meander_face = Param(pdt.TypeString, "Meander face on right side", "single", choices=["Single", "Two Face"])

    def build(self):
        launchers = self.produce_launchers("SMA8")
        self._produce_transmission_lines(launchers)

    def _produce_transmission_lines(self, launchers):
        distance = 700
        right_tr_x = 5000 + distance
        left_tr_x = 5000 - distance
        face1_box = self.get_box(1)

        # Left transmission line
        nodes = [Node(self.refpoints["NW_port"]),
                 Node(self.refpoints["NW_port_corner"] + pya.DPoint(0, - 2 * self.r)),
                 Node((left_tr_x, self.refpoints["NW_port_corner"].y - 2 * self.r)),
                 Node((left_tr_x, face1_box.p2.y), a=self.a_capped, b=self.b_capped),
                 Node((left_tr_x, face1_box.p1.y + 100), a=self.a, b=self.b),
                 Node((left_tr_x, self.refpoints["SW_port_corner"].y + 2 * self.r)),
                 Node(self.refpoints["SW_port_corner"] + pya.DPoint(0, 2 * self.r)),
                 Node(self.refpoints["SW_port"])
                 ]

        self.insert_cell(WaveguideComposite, nodes=nodes)

        # Right transmission line
        nodes = [Node(self.refpoints["NE_port"]),
                 Node(self.refpoints["NE_port_corner"] + pya.DPoint(0, - 2 * self.r)),
                 Node((right_tr_x, self.refpoints["NE_port_corner"].y - 2 * self.r)),
                 Node((right_tr_x, face1_box.p2.y),
                      a=self.a_capped, b=self.b_capped),
                 Node((right_tr_x, face1_box.p1.y + 100), a=self.a, b=self.b),
                 Node((right_tr_x, self.refpoints["SE_port_corner"].y + 2 * self.r)),
                 Node(self.refpoints["SE_port_corner"] + pya.DPoint(0, 2 * self.r)),
                 Node(self.refpoints["SE_port"])
                 ]

        self.insert_cell(WaveguideComposite, nodes=nodes)

        # Crossing transmission line
        nodes = [Node(self.refpoints["WN_port"]),
                 Node((face1_box.p1.x, self.refpoints["WN_port"].y),
                      a=self.a_capped, b=self.b_capped)]
        ref_x = left_tr_x
        ref_x_1 = ref_x - self.crossing_length / 2.
        ref_x_2 = ref_x + self.crossing_length / 2.

        last_y = launchers["WN"][0].y
        crossings = self.crossings  # must be even
        step = (launchers["WN"][0].y - launchers["WS"][0].y) / (crossings - 0.5) / 2
        wiggle = self.crossing_length / 2. + 250

        for i in range(crossings):
            nodes.append(Node((ref_x - wiggle, last_y)))
            nodes.append(Node((ref_x_1, last_y), face_id=self.face_ids[1]))
            nodes.append(Node((ref_x_2, last_y), face_id=self.face_ids[0]))
            nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x_2, last_y), face_id=self.face_ids[1]))
            nodes.append(Node((ref_x_1, last_y), face_id=self.face_ids[0]))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
        nodes.append(Node((face1_box.p1.x + 100, self.refpoints["WS_port"].y), a=self.a, b=self.b))
        nodes.append(Node(self.refpoints["WS_port_corner"]))
        nodes.append(Node(self.refpoints["WS_port"]))
        self.insert_cell(WaveguideComposite, nodes=nodes)

        # cross_talk
        ref_x = right_tr_x + self.cross_talk_distance + wiggle
        last_y = self.refpoints["EN_port"].y
        nodes = [Node(self.refpoints["EN_port"]),
                 Node(self.refpoints["EN_port_corner"]),
                 Node((face1_box.p2.x, self.refpoints["EN_port"].y),
                      a=self.a_capped, b=self.b_capped)]
        for i in range(crossings):
            if i == 0 and self.meander_face == "Two Face":
                nodes.append(
                    Node((ref_x + wiggle, last_y), face_id=self.face_ids[1]))
            else:
                nodes.append(Node((ref_x + wiggle, last_y)))
            nodes.append(Node((ref_x - wiggle, last_y)))
            last_y -= step
            nodes.append(Node((ref_x - wiggle, last_y)))
            if i == range(crossings)[-1] and self.meander_face == "Two Face":
                nodes.append(Node((ref_x + wiggle, last_y), face_id=self.face_ids[0]))
            else:
                nodes.append(Node((ref_x + wiggle, last_y)))
            last_y -= step
        nodes.append(Node((face1_box.p2.x - 100, self.refpoints["ES_port"].y),
                          a=self.a, b=self.b))
        nodes.append(Node(self.refpoints["ES_port_corner"]))
        nodes.append(Node(self.refpoints["ES_port"]))
        self.insert_cell(WaveguideComposite, nodes=nodes)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


"""This module is used for importing the KLayout Python API (pya).

Without this module, the API would need to be imported using ``import pya`` for usage in KLayout Editor and using
``import klayout.db`` for usage with standalone klayout package. To make it simple to create a python module for both
use cases, this module automatically imports the correct module and exposes it as ``pya``.

It also contains convenience functions to find the KLayout executable and the running session type.

Usage:
    from kqcircuits.pya_resolver import pya

"""

import os
import platform
from pathlib import Path
from shutil import which
try:
    import pya
    import pya as lay  # pylint: disable=unused-import
except ImportError:
    import klayout.db as pya
    from klayout import lay  # pylint: disable=unused-import


def is_standalone_session():
    return not hasattr(pya, 'Application')

def klayout_executable_command():
    """Returns the KLayout executable command's full path in the current OS. Or ``None`` if it is not found."""
    if is_standalone_session():
        if platform.system() == "Windows":
            klayout_path = which("klayout_app.exe")
            if klayout_path is None:  # try the default location
                dwp = Path(os.getenv("APPDATA")).joinpath("KLayout/klayout_app.exe")
                if dwp.is_file():
                    return str(dwp)
            return klayout_path
        klayout_path = which("klayout")  # Linux is simple :)
        if klayout_path is None and platform.system() == "Darwin":
            dwp = "/Applications/klayout.app/Contents/MacOS/klayout"
            if Path(dwp).is_file():
                return dwp
            dwp = "/Applications/KLayout/klayout.app/Contents/MacOS/klayout"
            if Path(dwp).is_file():
                return dwp
        return klayout_path
    else:  # The path of the currently running KLayout application
        return pya.Application.instance().applicationFilePath()



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from math import pi, cos, sin, atan
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.element import Element
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare


@add_parameters_from(FingerCapacitorSquare, "fixed_length", "a2", "b2", finger_width=10, finger_gap=5)
class SmoothCapacitor(Element):
    """The PCell declaration for a smooth finger capacitor.

    SmoothCapacitor is a finger capacitor, which has continuous geometry changes
    through the capacitance range. This leads to continuous capacitance function,
    which enables using capacitor inside numerical optimization methods.

    Capacitance range is achieved by changing single parameter called `finger_control`.
    """

    finger_control = Param(pdt.TypeDouble, "Continuously adjust finger number", 2.1,
               docstring="Parameter for capacitor growth (related to number of fingers per side)")
    ground_gap = Param(pdt.TypeDouble, "Gap between ground and finger", 10, unit="μm")

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def build(self):
        # constants
        r1 = self.finger_width / 2
        r2 = r1 + self.finger_gap
        scale = self.finger_width + self.finger_gap
        x_max = max(self.finger_control, 1.0 / self.finger_control) * scale - self.finger_gap / 2
        x_mid = x_max - self.finger_width
        xport = x_max + self.ground_gap

        def unit_vector(radians):
            return pya.DVector(cos(radians), sin(radians))

        def segment_points(start_pos, start_ang, turn, length):
            if length == 0:
                return []
            if turn == 0:
                return [start_pos + length * unit_vector(start_ang)]
            r = length / turn  # signed radius
            center = start_pos + r * unit_vector(start_ang + pi / 2)  # center of the turn circle
            num_pnts = max(round(abs(turn) * self.n / (2 * pi)), 1)  # number of new points
            return [center - r * unit_vector(start_ang + pi / 2 + turn * (n + 1) / num_pnts) for n in range(num_pnts)]

        def t_poly(bend, length):
            # Bottom 180-degree bend starting from origin
            ang = 3 * pi / 2
            pnts = segment_points(pya.DPoint(0, 0), ang, -pi, pi * r1)
            ang -= pi
            # Bend before straight segment
            bend0 = min(bend, pi / 2)
            pnts += segment_points(pnts[-1], ang, bend0, bend0 * r2)
            ang += bend0
            # Straight segment, 180-degree bend, and straight segment back
            pnts += segment_points(pnts[-1], ang, 0.0, length)
            pnts += segment_points(pnts[-1], ang, -pi, pi * r1)
            ang -= pi
            pnts += segment_points(pnts[-1], ang, 0.0, length)
            # Possible turn upwards if bend > pi/2
            turn = bend - pi / 2
            if turn > 0.0:
                pnts += segment_points(pnts[-1], ang, turn, turn * r2)
                ang += turn
            # The last bend back to origin
            last_bend = ang + pi / 2
            if last_bend > 1e-13:
                r3 = pnts[-1].x / (cos(last_bend) - 1)
                pnts += segment_points(pnts[-1], ang, -last_bend, last_bend * r3)
                pnts += segment_points(pnts[-1], -pi / 2, 0.0, max(pnts[-1].y, 0.0))
            return pya.DPolygon(pnts)

        def finger_polygon(order_number):
            if self.finger_control <= order_number:  # The finger does not exist for given order_number.
                return None
            trans = pya.DTrans(0, order_number % 2 == 1) * pya.DTrans(x_max, (order_number - 0.5) * scale)
            if self.finger_control <= 1.0:
                return t_poly(0.0, scale).transformed(trans)
            t_len = scale * pi/2  # length of 90-degree turn segment
            s_len = scale * (2 * self.finger_control - 3)  # length of straight segment
            f_len = s_len + 2 * t_len  # total length of finger (including two 90-degree turns and straight)
            x = (self.finger_control - order_number) * f_len
            if x < t_len:  # The first turn is not full 90 degrees.
                return t_poly((x / t_len) * pi / 2, 0.0).transformed(trans)
            if s_len < 0.0:  # The first turn is limited by finger length. This only happens when order_number=0.
                return t_poly(pi / 2 - 2 * atan(-s_len / scale), -s_len).transformed(trans)
            if x < t_len + s_len:  # The straight segment is not full length.
                return t_poly(pi / 2, x - t_len).transformed(trans)
            if x < 2 * f_len - t_len:  # The straight segment is full, but the last turn does not exist yet.
                return t_poly(pi / 2, s_len).transformed(trans)
            if x < 2 * f_len:  # The last turn is below 90 degrees.
                return t_poly(pi * (1 + (x - 2 * f_len) / (2 * t_len)), s_len).transformed(trans)
            return t_poly(pi, s_len).transformed(trans)

        def insert_wg_joint(reg, x0, xr, r):
            rr = r / self.layout.dbu
            reg += pya.Region(pya.DBox(xr, -r, 2*x0 - xr, r).to_itype(self.layout.dbu)).rounded_corners(rr, rr, self.n)
            reg -= pya.Region(pya.DBox(x0, -r, 2*x0 - xr, r).to_itype(self.layout.dbu))

        def middle_gap_fill():
            y = scale / 2
            x = (x_mid + x_max) / 2
            l = 2 * x if self.finger_control < 1 else scale
            rr = (self.finger_width / 2 + self.ground_gap) / self.layout.dbu
            return pya.Region(pya.DPolygon([pya.DPoint(-x, y), pya.DPoint(l - x, y),
                                            pya.DPoint(x, -y), pya.DPoint(x - l, -y)]).to_itype(self.layout.dbu)
                              ).sized(rr, 5).rounded_corners(rr, rr, self.n)

        def super_smoothen_region(reg, r):
            rr = r / self.layout.dbu
            reg_mod = reg.sized(rr, 5).sized(-rr, 5).rounded_corners(rr, 0, self.n).rounded_corners(0, rr, self.n)
            reg += reg_mod
            return reg.smoothed(1)

        # List of finger polygons
        i = 0
        polys = []
        while True:
            poly = finger_polygon(i)
            if poly is None:
                break
            polys.append(poly)
            i += 1

        # Create finger pad regions
        right_fingers = pya.Region([
            poly.to_itype(self.layout.dbu) for poly in polys
        ])
        left_fingers = right_fingers.transformed(pya.Trans(2))

        # Ground etch region
        region_ground = right_fingers + left_fingers
        region_ground.size(self.ground_gap / self.layout.dbu, 5)
        region_ground += middle_gap_fill()
        a2 = self.a if self.a2 < 0 else self.a2
        b2 = self.b if self.b2 < 0 else self.b2
        insert_wg_joint(region_ground, xport, x_mid - self.ground_gap, b2 + a2/2)
        insert_wg_joint(region_ground, -xport, -x_mid + self.ground_gap, self.b + self.a/2)
        region_ground = super_smoothen_region(region_ground, self.finger_gap + self.ground_gap)

        # Finalize finger pad regions
        insert_wg_joint(right_fingers, xport, x_mid, a2/2)
        insert_wg_joint(left_fingers, -xport, -x_mid, self.a/2)
        right_fingers = super_smoothen_region(right_fingers, self.finger_gap)
        left_fingers = super_smoothen_region(left_fingers, self.finger_gap)

        # Insert waveguide segments in both ends, if fixed_length is set
        if self.fixed_length != 0:
            xfixed = self.fixed_length / 2
            if xfixed < xport:
                raise ValueError(f"SmoothCapacitor parameters not compatible with fixed_length={self.fixed_length}")
            region_ground += pya.Region(pya.DBox(xport, -b2 - a2/2, xfixed, b2 + a2/2).to_itype(self.layout.dbu)) + \
                pya.Region(pya.DBox(-xfixed, -self.b - self.a/2, -xport, self.b + self.a/2).to_itype(self.layout.dbu))
        else:
            xfixed = xport
        # Always insert tolerance to secure trace connection
        right_fingers += pya.Region(pya.DBox(xport-0.001, -a2/2, xfixed+1, a2/2).to_itype(self.layout.dbu))
        left_fingers += pya.Region(pya.DBox(-xfixed-1, -self.a/2, -xport+0.001, self.a/2).to_itype(self.layout.dbu))
        xport = xfixed

        # Create shapes into cell
        region = region_ground - right_fingers - left_fingers
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region)

        # protection
        region_protection = region_ground.sized(self.margin / self.layout.dbu, 5)
        self.add_protection(region_protection)

        # Add size into annotation layer
        self.cell.shapes(self.get_layer("annotations")).insert(pya.DText(str(round(self.finger_control, 5)), 0, 0))

        # Create ports
        self.add_port("a", pya.DPoint(-xport, 0), pya.DVector(-1, 0))
        self.add_port("b", pya.DPoint(xport, 0), pya.DVector(1, 0))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.elements.element import Element
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare


@add_parameters_from(FingerCapacitorSquare, a2=Element.a*2, b2=Element.b*2)
class WaveguideCoplanarTaper(Element):
    """The PCell declaration of a taper segment of a coplanar waveguide.

    .. MARKERS_FOR_PNG 0,0,31.2,0 0,5,0,-5 31.2,-10,31.2,10
    """

    taper_length = Param(pdt.TypeDouble, "Taper length", 10 * math.pi, unit="μm")
    m2 = Param(pdt.TypeDouble, "Margin of right waveguide protection layer", 5 * 2, unit="μm")

    def build(self):
        #
        # gap 1
        pts = [
            pya.DPoint(0, self.a / 2 + 0),
            pya.DPoint(self.taper_length, self.a2 / 2 + 0),
            pya.DPoint(self.taper_length, self.a2 / 2 + self.b2),
            pya.DPoint(0, self.a / 2 + self.b)
        ]
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        # gap 2
        pts = [
            pya.DPoint(0, -self.a / 2 + 0),
            pya.DPoint(self.taper_length, -self.a2 / 2 + 0),
            pya.DPoint(self.taper_length, -self.a2 / 2 - self.b2),
            pya.DPoint(0, -self.a / 2 - self.b)
        ]
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        # Protection layer
        pts = [
            pya.DPoint(0, -self.a / 2 - self.b - self.margin),
            pya.DPoint(self.taper_length, -self.a2 / 2 - self.b2 - self.m2),
            pya.DPoint(self.taper_length, self.a2 / 2 + self.b2 + self.m2),
            pya.DPoint(0, self.a / 2 + self.b + self.margin)
        ]
        self.add_protection(pya.DPolygon(pts))
        # Waveguide layer
        pts = [
            pya.DPoint(0, self.a / 2),
            pya.DPoint(self.taper_length, self.a2 / 2),
            pya.DPoint(self.taper_length, -self.a2 / 2),
            pya.DPoint(0, -self.a / 2)
        ]
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("waveguide_path")).insert(shape)
        pts = [
            pya.DPoint(0, 0),
            pya.DPoint(self.taper_length, 0),
        ]
        shape = pya.DPath(pts, min(self.a, self.a2))
        self.cell.shapes(self.get_layer("waveguide_path")).insert(shape)
        # refpoints for connecting to waveguides
        self.add_port("a", pya.DPoint(0, 0), pya.DVector(-1, 0))
        self.add_port("b", pya.DPoint(self.taper_length, 0), pya.DVector(1, 0))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from autologging import logged

from kqcircuits.pya_resolver import pya
from kqcircuits.elements.fluxlines.fluxline import Fluxline


@logged
class FluxlineStandard(Fluxline):
    """Fluxline variant "standard".

     .. MARKERS_FOR_PNG 9,-21 11,-17 -10,-9.7
    """

    def build(self):
        # shorthands
        a = self.a  # waveguide center width
        b = self.b  # waveguide gap width
        fb = self.fluxline_gap_width  # fluxline gap width
        fa = self.fluxline_gap_width * (self.a / self.b)  # fluxline center width
        w = self.fluxline_width
        l1 = 30  # length of straight down section
        l2 = 50  # length of taper to waveguide port

        # origin at edge of the qubit gap
        right_gap = pya.DPolygon([
            pya.DPoint(-w / 2 - fa / 2, -fa),
            pya.DPoint(w / 2 + fa / 2 + fb, -fa),
            pya.DPoint(w / 2 + fa / 2 + fb, -fa - l1),
            pya.DPoint(a / 2 + b, -fa - l1 - l2),
            pya.DPoint(a / 2, -fa - l1 - l2),
            pya.DPoint(w / 2 + fa / 2, -fa - l1),
            pya.DPoint(w / 2 + fa / 2, -fa - fb),
            pya.DPoint(-w / 2 - fa / 2, -fa - fb)
        ])
        left_gap = pya.DPolygon([
            pya.DPoint(-w / 2 - fa / 2, -2 * fa - fb),
            pya.DPoint(w / 2 - fa / 2, -2 * fa - fb),
            pya.DPoint(w / 2 - fa / 2, -fa - l1),
            pya.DPoint(-a / 2, -fa - l1 - l2),
            pya.DPoint(-a / 2 - b, -fa - l1 - l2),
            pya.DPoint(w / 2 - fa / 2 - fb, -fa - l1),
            pya.DPoint(w / 2 - fa / 2 - fb, -2 * fa - 2 * fb),
            pya.DPoint(-w / 2 - fa / 2, -2 * fa - 2 * fb)
        ])

        self._insert_fluxline_shapes(left_gap, right_gap)
        self._add_fluxline_refpoints(pya.DPoint(0, -fa - l1 - l2))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.elements.chip_frame import ChipFrame
from kqcircuits.chips.chip import Chip
from kqcircuits.pya_resolver import pya
from kqcircuits.test_structures.junction_test_pads.junction_test_pads import JunctionTestPads
from kqcircuits.junctions.junction import Junction


@add_parameters_from(JunctionTestPads, "*", "loop_area", "junction_width", pad_spacing=200)
@add_parameters_from(Junction, "junction_type")
@add_parameters_from(ChipFrame, "marker_types")
class JunctionTest2(Chip):
    """The PCell declaration for a JunctionTest2 chip."""

    pad_width = Param(pdt.TypeDouble, "Pad Width", 500, unit="μm")
    junctions_horizontal = Param(pdt.TypeBoolean, "Horizontal (True) or vertical (False) junctions", True)
    small_loop_area = Param(pdt.TypeDouble, "Test SQUIDs small loop area",
                                 default=80, unit="μm")
    large_loop_area = Param(pdt.TypeDouble, "Test SQUIDs large loop area",
                                 default=130, unit="μm")
    junction_width_small = Param(pdt.TypeDouble, "Test SQUIDs Junction finger width starting value (small loop)",
                                      default=0.15, unit="μm")
    junction_width_large = Param(pdt.TypeDouble, "Test SQUIDs Junction finger width starting value (large loop)",
                                      default=0.08, unit="μm")
    junction_width_step_increment_small = Param(pdt.TypeDouble, "Junction finger width step increment (small loop)",
                                                default=0.01, unit="μm")
    junction_width_step_increment_large = Param(pdt.TypeDouble, "Junction finger width step increment (large loop)",
                                                default=0.03, unit="μm")
    pads_loop = Param(pdt.TypeList, "Select large or small loop area for each central test pad",
                      default=["large", "large", "small", "small", "small", "large"])

    def coerce_parameters_impl(self):
        self.sync_parameters(JunctionTestPads)

    def build(self):
        left = self.box.left
        right = self.box.right
        top = self.box.top

        arrays_coordinates = [(left + 300, top - 2000 - 6000), (left + 2000, top - 300 - 9400),
                                   (left + 2000 + 2*(self.pad_spacing+ self.pad_width), top - 300 - 9400),
                                   (left + 2000 + 4*(self.pad_spacing + self.pad_width),
                                    top - 300 - 9400), (left + 2000 + 6*(self.pad_spacing + self.pad_width), top - 300
                                    - 9400), (right - 300 - 1700, top - 2000 - 6000)]
        area_height = [6000, 9400, 9400, 9400, 9400, 6000]
        squid_indexing_small = 0
        squid_indexing_large = 0
        name = "testarray"

        for j,array_coordinates in enumerate(arrays_coordinates):

            squids_per_test_array = area_height[j] // (self.pad_spacing + self.pad_width)

            if self.pads_loop[j] == "small":
                junction_width_starting_value = self.junction_width_small + squid_indexing_small * \
                                                self.junction_width_step_increment_small

                squid_indexing_small += squids_per_test_array
                junction_width_steps = [junction_width_starting_value, self.junction_width_step_increment_small]
                loop_area = self.small_loop_area

            else:
                junction_width_starting_value = self.junction_width_large + squid_indexing_large * \
                                                self.junction_width_step_increment_large

                squid_indexing_large += squids_per_test_array
                junction_width_steps = [junction_width_starting_value, self.junction_width_step_increment_large]
                loop_area = self.large_loop_area

            cell = self.add_element(JunctionTestPads,
                                margin=20,
                                area_height=area_height[j],
                                area_width=3*self.pad_spacing + 2*self.pad_width,
                                pad_width=self.pad_width,
                                junction_width_steps=junction_width_steps,
                                loop_area =loop_area,
                                only_arms=True,
                                )
            self.insert_cell(cell, pya.DTrans(0, False, array_coordinates[0], array_coordinates[1]),
                             name + "_{}".format(j+1))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
DPoint = pya.DPoint


class Port:
    """Base data structure for simulation ports.

    Depending on your simulation type, these produce excitations, set potentials, or act as ideal RLC lumped elements.
    """
    def __init__(self, number: int,
                 resistance: float = 50, reactance: float = 0, inductance: float = 0, capacitance: float = 0,
                 face: int = 0, junction: bool = False):
        """
        Args:
            number: Port number.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
        """
        self.number = number
        self.resistance = resistance
        self.reactance = reactance
        self.inductance = inductance
        self.capacitance = capacitance
        self.face = face
        self.junction = junction
        self.type = type(self).__name__

    def as_dict(self):
        """Returns attributes as a dictionary."""
        return vars(self)


class InternalPort(Port):
    """Data structure for ports inside the simulation area."""
    def __init__(self, number: int, signal_location: DPoint, ground_location: DPoint = None,
                 resistance: float = 50, reactance: float = 0, inductance: float = 0, capacitance: float = 0,
                 face: int = 0, junction: bool = False, signal_layer: str = 'signal'):
        """
        Args:
            number: Port number.
            signal_location: Edge location for signal source.
            ground_location: Edge location to connect signal to. Usually ground.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
            signal_layer: Manual override for simulation signal layer.
                May be used to set ports across the ground layer with ``ground``.
        """
        super().__init__(number, resistance, reactance, inductance, capacitance, face, junction)
        self.signal_location = signal_location
        self.signal_layer = signal_layer
        if ground_location is not None:
            self.ground_location = ground_location


class EdgePort(Port):
    """Data structure for ports at the edge of the simulation area."""
    def __init__(self, number: int, signal_location: DPoint,
                 resistance: float = 50, reactance: float = 0, inductance: float = 0, capacitance: float = 0,
                 deembed_len: float = None, face: int = 0, junction: bool = False):
        """
        Args:
            number: Port number.
            signal_location: Edge location for signal source.
            resistance: Real part of impedance. Given in Ohms (:math:`\\Omega`).
            reactance: Imaginary part of impedance. Given in Ohms (:math:`\\Omega`).
            inductance: Inductance of the element. Given in Henrys (:math:`\\text{H}`).
            capacitance: Capacitance of the element. Given in Farads (:math:`\\text{F}`).
            deembed_len: Port de-embedding length. Given in simulation units, usually microns (:math:`\\text{um}`).
            face: Integer-valued face index for the port.
            junction: Whether this port models a SQUID/Junction. Used in EPR calculations.
        """
        super().__init__(number, resistance, reactance, inductance, capacitance, face, junction)
        self.signal_location = signal_location
        self.deembed_len = deembed_len
